from mccode_antlr.loader.loader import parse_mcstas_instr
from textwrap import dedent

from .compiled import compiled, compile_and_run


@compiled
def test_without_components():
    instr = parse_mcstas_instr(dedent("""\
    DEFINE INSTRUMENT without_components(int dummy=0)
    DECLARE %{
    int i_declare;
    %}
    INITIALIZE %{
    i_declare = 1;
    printf("Ran without components!\\ndummy=%d\\n", dummy);
    %}
    TRACE
    FINALLY %{ %}
    END
    """))
    results, files = compile_and_run(instr, '-n 1 dummy=101')
    lines = results.decode('utf-8').splitlines()
    assert len(lines) == 3
    assert lines[0] == "Ran without components!"
    assert lines[1] == "dummy=101"



@compiled
def test_template_instr():
    """
    This test failed previously because its name is a reserved word in C++ (template)
    """
    instr = parse_mcstas_instr(dedent("""\
    DEFINE INSTRUMENT template(Par1=1)
    DECLARE %{ %}
    INITIALIZE %{ %}
    TRACE
    COMPONENT Origin = Progress_bar() AT (0,0,0) ABSOLUTE

    FINALLY %{ %}
    END
    """))
    compile_and_run(instr, "-n 1 Par1=2")


@compiled
def test_component_declare_variable_initialised():
    instr = parse_mcstas_instr(dedent("""\
    DEFINE INSTRUMENT namedsomething(dummy=0)
    DECLARE %{
    int initialized;
    %}
    INITIALIZE %{
    initialized = -1;
    %}
    TRACE
    component Origin = Progress_bar() at (0, 0, 0) absolute
    component mirror = SupermirrorFlat() at (0, 0, 0) absolute
    FINALLY %{ %}
    END
    """))
    compile_and_run(instr, '-n 1 dummy=1')


@compiled
def test_function_pointer_declare_parameter():
    instr = parse_mcstas_instr(dedent("""\
    DEFINE INSTRUMENT with_function_pointers(int which=0)
    DECLARE %{
    int add(int a, int b){ return a + b; }
    int subtract(int a, int b){ return a - b; }
    int multiply(int a, int b){ return a * b; }
    
    int (*fun_ptr)(int, int);
    int (*fun_ptr_arr[])(int, int) = {add, subtract, multiply};
    %}
    INITIALIZE %{
    switch (which){
    case 0: fun_ptr = add; break;
    case 1: fun_ptr = subtract; break;
    case 2: fun_ptr = multiply; break;
    default:
        printf("Valid settings for 'which' are {0, 1, 2}. 'which=%d' is invalid.\\n", which);
        exit(-1);
    }
    %}
    TRACE
    COMPONENT first = Arm() AT (0, 0, 0) ABSOLUTE EXTEND %{
        printf("first=%d\\n", fun_ptr(2, 1));
    %}
    COMPONENT second = Arm() AT (0, 0, 0) ABSOLUTE EXTEND %{
        int choice = instrument->_parameters.which; // _should_ this be necessary?
        printf("second=%d\\n", fun_ptr_arr[choice](2, 1));
    %} 
    FINALLY %{ %}
    END
    """))
    for which in (0, 1, 2):
        results, files = compile_and_run(instr, f'-n 1 which={which}')
        lines = results.decode('utf-8').splitlines()
        res = None
        if which == 0:
            res = 2 + 1
        elif which == 1:
            res = 2 - 1
        elif which == 2:
            res = 2 * 1
        assert lines[0] == f"first={res}"
        assert lines[1] == f"second={res}"


@compiled
def test_function_pointer_component_declare_parameter():
    from mccode_antlr.reader.registry import InMemoryRegistry
    in_memory_registry = InMemoryRegistry('test_components')
    comp_name = 'declares_function_pointer'
    in_memory_registry.add_comp(comp_name, dedent(rf"""
    DEFINE COMPONENT {comp_name} DEFINITION PARAMETERS ()
    SETTING PARAMETERS (int selector=0, int A=0, int B=0)
    OUTPUT PARAMETERS ()
    SHARE %{{
        int add(int a, int b){{return a + b;}}
        int sub(int a, int b){{return a - b;}}
        int mul(int a, int b){{return a * b;}}
    %}}
    DECLARE %{{
        int (*fun_ptr)(int, int);
        int (*fun_ptr_arr[])(int, int) = {{add, sub, mul}};
    %}}
    INITIALIZE %{{
        switch (selector) {{
        case 0: fun_ptr = add; break;
        case 1: fun_ptr = sub; break;
        case 2: fun_ptr = mul; break;
        default:
        printf("%s: Invalid selector=%d, valid settings are {{0, 1, 2}}\n", NAME_CURRENT_COMP, selector);
        exit(-1);
        }}
    %}}
    TRACE %{{
        printf("%s=%d\n", NAME_CURRENT_COMP, fun_ptr(A, B));
        printf("%s=%d\n", NAME_CURRENT_COMP, fun_ptr_arr[selector](A, B));
    %}}
    SAVE %{{
    %}}
    FINALLY %{{
    %}}
    END
    """))

    instr = parse_mcstas_instr(dedent(rf"""
    DEFINE INSTRUMENT with_component_function_pointers(int which=0)
    DECLARE %{{ %}}
    INITIALIZE %{{ %}}
    TRACE
    COMPONENT first = {comp_name}(selector=which, A=2, B=1) AT (0, 0, 0) ABSOLUTE
    COMPONENT second = {comp_name}(selector=which, A=1, B=3) AT (0, 0, 0) ABSOLUTE
    FINALLY %{{ %}}
    END
    """), registries=[in_memory_registry])
    for which in (0, 1, 2):
        results, files = compile_and_run(instr, f'-n 1 which={which}')
        lines = results.decode('utf-8').splitlines()
        print(lines)
        a, b = 0, 0
        if which == 0:
            a, b = 2 + 1, 1 + 3
        elif which == 1:
            a, b = 2 - 1, 1 - 3
        elif which == 2:
            a, b = 2 * 1, 1 * 3
        assert lines[0] == f"first={a}"
        assert lines[1] == f"first={a}"
        assert lines[2] == f"second={b}"
        assert lines[3] == f"second={b}"


@compiled
def test_struct_instance_parameter():
    instr = parse_mcstas_instr(dedent(r"""
    DEFINE INSTRUMENT templateTAS(DM=3.3539)
    DECLARE
    %{
      struct machine_hkl_struct {
        double dm;
        double refl;
        char refl_file[100];
      } machine_hkl;
    %}
    INITIALIZE
    %{
    machine_hkl.dm = DM; 
    machine_hkl.refl = 0.7;
    memset(machine_hkl.refl_file, '\0', sizeof(machine_hkl.refl_file));
    if (fabs(machine_hkl.dm - 3.355) < 0.2) {
      machine_hkl.refl = 1.0;
      strcpy(machine_hkl.refl_file, "HOPG.rfl");
    }
    %}
    TRACE
    COMPONENT Origin=Progress_bar()
    AT (0,0,0) ABSOLUTE
    
    COMPONENT PG1Xtal = Monochromator_curved(
      width  = 0.1, height = 0.1, NH=1, NV=1, RV=0, RH=0,
      DM = machine_hkl.dm, r0 = machine_hkl.refl, reflect = machine_hkl.refl_file)
    AT (0, 0, 0) RELATIVE Origin
    EXTEND
    %{
    printf("PG1Xtal.r0=%3.1f, PG1Xtal.reflect='%s'\n", r0, reflect);
    %}

    COMPONENT PG2Xtal = Monochromator_curved(
      width  = 0.1, height = 0.1, NH=1, NV=1, RV=0, RH=0,
      DM = machine_hkl.dm, 
      r0 = (fabs(machine_hkl.dm-3.355) < 0.2 ? 1: 0.7), 
      reflect=(fabs(machine_hkl.dm-3.355) < 0.2 ? "HOPG.rfl" : ""))
    AT (0, 0, 0) RELATIVE Origin
    EXTEND
    %{
    printf("PG2Xtal.r0=%3.1f, PG2Xtal.reflect='%s'\n", r0, reflect);
    %}

    END
    """))
    for dm, r0, filename in ((1.55, 0.7, ''), (3.355, 1.0, 'HOPG.rfl')):
        results, files = compile_and_run(instr, f'-n 1 DM={dm}')
        lines = results.decode('utf-8').splitlines()
        print('\n'.join(lines))
        assert lines[0] == '[templateTAS] Initialize'
        assert lines[1] == f"PG1Xtal.r0={r0:3.1f}, PG1Xtal.reflect='{filename}'"
        assert lines[2] == f"PG2Xtal.r0={r0:3.1f}, PG2Xtal.reflect='{filename}'"


@compiled
def test_copy_extend_instance_parameter():
    from mccode_antlr.reader.registry import InMemoryRegistry
    in_memory_registry = InMemoryRegistry('test_components')
    comp_name = 'has_parameter'
    in_memory_registry.add_comp(comp_name, dedent(rf"""
    DEFINE COMPONENT {comp_name} DEFINITION PARAMETERS ()
    SETTING PARAMETERS (int parameter=1)
    TRACE %{{
        printf("%s=%d\n", NAME_CURRENT_COMP, parameter);
    %}}
    END
    """))

    instr = parse_mcstas_instr(dedent(rf"""
    DEFINE INSTRUMENT with_copied_then_extended_instance(int which=0)
    TRACE
    COMPONENT p1 = {comp_name}() AT (0, 0, 0) ABSOLUTE
    COMPONENT p2 = {comp_name}(parameter=2) AT (0, 0, 0) ABSOLUTE
    EXTEND %{{
      printf("%s=%d\n", NAME_CURRENT_COMP, parameter+10);
    %}}
    COMPONENT p3 = COPY(p1) AT (0, 0, 0) ABSOLUTE
    COMPONENT p4 = COPY(p2)(parameter=4) AT (0, 0, 0) ABSOLUTE
    COMPONENT COPY(p1) = COPY(p3)(parameter=5) AT (0, 0, 0) ABSOLUTE
    COMPONENT COPY(p4) = COPY(p4) AT (0, 0, 0) ABSOLUTE
    EXTEND %{{
      printf("%s=%d\n", NAME_CURRENT_COMP, parameter+100);
    %}}
    END
    """), registries=[in_memory_registry])

    results, files = compile_and_run(instr, f'-n 1 which=1')
    lines = results.decode('utf-8').splitlines()
    print(lines)
    name_parameter = (('p1', 1), ('p2', 2), ('p2', 12), ('p3', 1),
                      ('p4', 4), ('p4', 14), ('p1_5', 5),
                      ('p4_6', 4), ('p4_6', 104))
    for (name, parameter), line in zip(name_parameter, lines):
        assert line == f"{name}={parameter}"


from __future__ import annotations

import types
from antlr4 import InputStream, CommonTokenStream
from antlr4.tree.Tree import ParseTree
from antlr4.error.ErrorListener import ErrorListener


def parse(
        lexer_class,
        parser_class,
        stream:InputStream,
        entry_rule_name:str,
        error_listener: ErrorListener | None = None
) -> ParseTree:
    lexer = lexer_class(stream)
    if error_listener is not None:
        lexer.removeErrorListeners()
        lexer.addErrorListener(error_listener)
    token_stream = CommonTokenStream(lexer)

    parser = parser_class(token_stream)
    if error_listener is not None:
        parser.removeErrorListeners()
        parser.addErrorListener(error_listener)

    entry_rule_func = getattr(parser, entry_rule_name, None)
    if not isinstance(entry_rule_func, types.MethodType):
        raise ValueError("Invalid entry_rule_name '%s'" % entry_rule_name)
    return entry_rule_func()


from dataclasses import dataclass, field
from .instance import Instance

@dataclass
class Group:
    name: str
    index: int
    ids: list[int] = field(default_factory=list)
    members: list[Instance] = field(default_factory=list)

    def add(self, id: int, inst: Instance):
        self.ids.append(id)
        self.members.append(inst)

    def is_leader(self, index: int):
        return len(self.ids) and self.ids[0] == index

    @property
    def first_id(self):
        return min(self.ids)

    @property
    def last_id(self):
        return max(self.ids)

    @property
    def first(self):
        fid = self.first_id
        return [m for i, m in zip(self.ids, self.members) if i == fid][0]

    @property
    def last(self):
        lid = self.last_id
        return [m for i, m in zip(self.ids, self.members) if i == lid][0]

    def copy(self):
        return Group(self.name, self.index, self.ids.copy(), self.members.copy())



from .loader import parse_mcstas_instr, parse_mcxtrace_instr, load_mcstas_instr, load_mcxtrace_instr
from .datfile import read_mccode_dat, DatFile1D, DatFile2D, write_combined_mccode_dats
from .simfile import read_mccode_sim, combine_mccode_sims, write_combined_mccode_sims

__all__ = [
    'parse_mcstas_instr',
    'parse_mcxtrace_instr',
    'load_mcstas_instr',
    'load_mcxtrace_instr',
    'read_mccode_dat',
    'write_combined_mccode_dats',
    'read_mccode_sim',
    'combine_mccode_sims',
    'write_combined_mccode_sims',
    'DatFile2D',
    'DatFile1D'
]



import unittest


class MetadataTestCase(unittest.TestCase):
    def test_direct_creation(self):
        from mccode_antlr.common import MetaData
        md = MetaData.from_instance_tokens('instance_source', 'mimetype', 'metadata_name', 'metadata_value')
        self.assertEqual(md.source.name, 'instance_source')
        self.assertEqual(md.source.type_name, 'Instance')
        self.assertEqual(md.name, 'metadata_name')
        self.assertEqual(md.mimetype, 'mimetype')
        self.assertEqual(md.value, 'metadata_value')

    def test_parsed(self):
        from mccode_antlr.loader import parse_mcstas_instr
        from json import loads
        instr = """
        DEFINE INSTRUMENT splitRunTest(a1=0, a2=0, virtual_source_x=0.05, virtual_source_y=0.1, string newname)
        TRACE
        COMPONENT origin = Arm() AT (0, 0, 0) ABSOLUTE
        COMPONENT point = Arm() AT (0, 0, 0.8) RELATIVE origin
        METADATA "application/text" "names with spaces keep their quotes" %{
            This is some unparsed metadata that will be included as a literal string in the instrument.
        %}
        COMPONENT line = Arm() AT (0, 0, 1) RELATIVE PREVIOUS METADATA "application/json" identifier_name %{
            {"key": "value", "array": [1, 2, 3]}
        %}
        END
        """
        instr = parse_mcstas_instr(instr)
        self.assertEqual(len(instr.components), 3)
        self.assertEqual(instr.components[1].name, 'point')
        self.assertEqual(len(instr.components[1].metadata), 1)
        md = instr.components[1].metadata[0]
        self.assertEqual(md.source.name, 'point')
        self.assertEqual(md.source.type_name, 'Instance')
        self.assertEqual(md.name, '"names with spaces keep their quotes"')
        self.assertEqual(md.mimetype, 'application/text')
        self.assertEqual(md.value, """
            This is some unparsed metadata that will be included as a literal string in the instrument.
        """)
        self.assertEqual(instr.components[2].name, 'line')
        self.assertEqual(len(instr.components[2].metadata), 1)
        md = instr.components[2].metadata[0]
        self.assertEqual(md.source.name, 'line')
        self.assertEqual(md.source.type_name, 'Instance')
        self.assertEqual(md.name, 'identifier_name')
        self.assertEqual(md.mimetype, 'application/json')
        self.assertEqual(loads(md.value), {'key': 'value', 'array': [1, 2, 3]})



if __name__ == '__main__':
    unittest.main()



# Generated from /home/g/PycharmProjects/mccode-antlr/src/grammar/C.g4 by ANTLR 4.13.2
from antlr4 import *
if "." in __name__:
    from .CParser import CParser
else:
    from CParser import CParser

# This class defines a complete listener for a parse tree produced by CParser.
class CListener(ParseTreeListener):

    # Enter a parse tree produced by CParser#functionCall.
    def enterFunctionCall(self, ctx:CParser.FunctionCallContext):
        pass

    # Exit a parse tree produced by CParser#functionCall.
    def exitFunctionCall(self, ctx:CParser.FunctionCallContext):
        pass


    # Enter a parse tree produced by CParser#primaryExpression.
    def enterPrimaryExpression(self, ctx:CParser.PrimaryExpressionContext):
        pass

    # Exit a parse tree produced by CParser#primaryExpression.
    def exitPrimaryExpression(self, ctx:CParser.PrimaryExpressionContext):
        pass


    # Enter a parse tree produced by CParser#genericSelection.
    def enterGenericSelection(self, ctx:CParser.GenericSelectionContext):
        pass

    # Exit a parse tree produced by CParser#genericSelection.
    def exitGenericSelection(self, ctx:CParser.GenericSelectionContext):
        pass


    # Enter a parse tree produced by CParser#genericAssocList.
    def enterGenericAssocList(self, ctx:CParser.GenericAssocListContext):
        pass

    # Exit a parse tree produced by CParser#genericAssocList.
    def exitGenericAssocList(self, ctx:CParser.GenericAssocListContext):
        pass


    # Enter a parse tree produced by CParser#genericAssociation.
    def enterGenericAssociation(self, ctx:CParser.GenericAssociationContext):
        pass

    # Exit a parse tree produced by CParser#genericAssociation.
    def exitGenericAssociation(self, ctx:CParser.GenericAssociationContext):
        pass


    # Enter a parse tree produced by CParser#postfixExpression.
    def enterPostfixExpression(self, ctx:CParser.PostfixExpressionContext):
        pass

    # Exit a parse tree produced by CParser#postfixExpression.
    def exitPostfixExpression(self, ctx:CParser.PostfixExpressionContext):
        pass


    # Enter a parse tree produced by CParser#argumentExpressionList.
    def enterArgumentExpressionList(self, ctx:CParser.ArgumentExpressionListContext):
        pass

    # Exit a parse tree produced by CParser#argumentExpressionList.
    def exitArgumentExpressionList(self, ctx:CParser.ArgumentExpressionListContext):
        pass


    # Enter a parse tree produced by CParser#unaryExpression.
    def enterUnaryExpression(self, ctx:CParser.UnaryExpressionContext):
        pass

    # Exit a parse tree produced by CParser#unaryExpression.
    def exitUnaryExpression(self, ctx:CParser.UnaryExpressionContext):
        pass


    # Enter a parse tree produced by CParser#unaryOperator.
    def enterUnaryOperator(self, ctx:CParser.UnaryOperatorContext):
        pass

    # Exit a parse tree produced by CParser#unaryOperator.
    def exitUnaryOperator(self, ctx:CParser.UnaryOperatorContext):
        pass


    # Enter a parse tree produced by CParser#castExpression.
    def enterCastExpression(self, ctx:CParser.CastExpressionContext):
        pass

    # Exit a parse tree produced by CParser#castExpression.
    def exitCastExpression(self, ctx:CParser.CastExpressionContext):
        pass


    # Enter a parse tree produced by CParser#multiplicativeExpression.
    def enterMultiplicativeExpression(self, ctx:CParser.MultiplicativeExpressionContext):
        pass

    # Exit a parse tree produced by CParser#multiplicativeExpression.
    def exitMultiplicativeExpression(self, ctx:CParser.MultiplicativeExpressionContext):
        pass


    # Enter a parse tree produced by CParser#additiveExpression.
    def enterAdditiveExpression(self, ctx:CParser.AdditiveExpressionContext):
        pass

    # Exit a parse tree produced by CParser#additiveExpression.
    def exitAdditiveExpression(self, ctx:CParser.AdditiveExpressionContext):
        pass


    # Enter a parse tree produced by CParser#shiftExpression.
    def enterShiftExpression(self, ctx:CParser.ShiftExpressionContext):
        pass

    # Exit a parse tree produced by CParser#shiftExpression.
    def exitShiftExpression(self, ctx:CParser.ShiftExpressionContext):
        pass


    # Enter a parse tree produced by CParser#relationalExpression.
    def enterRelationalExpression(self, ctx:CParser.RelationalExpressionContext):
        pass

    # Exit a parse tree produced by CParser#relationalExpression.
    def exitRelationalExpression(self, ctx:CParser.RelationalExpressionContext):
        pass


    # Enter a parse tree produced by CParser#equalityExpression.
    def enterEqualityExpression(self, ctx:CParser.EqualityExpressionContext):
        pass

    # Exit a parse tree produced by CParser#equalityExpression.
    def exitEqualityExpression(self, ctx:CParser.EqualityExpressionContext):
        pass


    # Enter a parse tree produced by CParser#andExpression.
    def enterAndExpression(self, ctx:CParser.AndExpressionContext):
        pass

    # Exit a parse tree produced by CParser#andExpression.
    def exitAndExpression(self, ctx:CParser.AndExpressionContext):
        pass


    # Enter a parse tree produced by CParser#exclusiveOrExpression.
    def enterExclusiveOrExpression(self, ctx:CParser.ExclusiveOrExpressionContext):
        pass

    # Exit a parse tree produced by CParser#exclusiveOrExpression.
    def exitExclusiveOrExpression(self, ctx:CParser.ExclusiveOrExpressionContext):
        pass


    # Enter a parse tree produced by CParser#inclusiveOrExpression.
    def enterInclusiveOrExpression(self, ctx:CParser.InclusiveOrExpressionContext):
        pass

    # Exit a parse tree produced by CParser#inclusiveOrExpression.
    def exitInclusiveOrExpression(self, ctx:CParser.InclusiveOrExpressionContext):
        pass


    # Enter a parse tree produced by CParser#logicalAndExpression.
    def enterLogicalAndExpression(self, ctx:CParser.LogicalAndExpressionContext):
        pass

    # Exit a parse tree produced by CParser#logicalAndExpression.
    def exitLogicalAndExpression(self, ctx:CParser.LogicalAndExpressionContext):
        pass


    # Enter a parse tree produced by CParser#logicalOrExpression.
    def enterLogicalOrExpression(self, ctx:CParser.LogicalOrExpressionContext):
        pass

    # Exit a parse tree produced by CParser#logicalOrExpression.
    def exitLogicalOrExpression(self, ctx:CParser.LogicalOrExpressionContext):
        pass


    # Enter a parse tree produced by CParser#conditionalExpression.
    def enterConditionalExpression(self, ctx:CParser.ConditionalExpressionContext):
        pass

    # Exit a parse tree produced by CParser#conditionalExpression.
    def exitConditionalExpression(self, ctx:CParser.ConditionalExpressionContext):
        pass


    # Enter a parse tree produced by CParser#assignmentExpression.
    def enterAssignmentExpression(self, ctx:CParser.AssignmentExpressionContext):
        pass

    # Exit a parse tree produced by CParser#assignmentExpression.
    def exitAssignmentExpression(self, ctx:CParser.AssignmentExpressionContext):
        pass


    # Enter a parse tree produced by CParser#assignmentOperator.
    def enterAssignmentOperator(self, ctx:CParser.AssignmentOperatorContext):
        pass

    # Exit a parse tree produced by CParser#assignmentOperator.
    def exitAssignmentOperator(self, ctx:CParser.AssignmentOperatorContext):
        pass


    # Enter a parse tree produced by CParser#expression.
    def enterExpression(self, ctx:CParser.ExpressionContext):
        pass

    # Exit a parse tree produced by CParser#expression.
    def exitExpression(self, ctx:CParser.ExpressionContext):
        pass


    # Enter a parse tree produced by CParser#constantExpression.
    def enterConstantExpression(self, ctx:CParser.ConstantExpressionContext):
        pass

    # Exit a parse tree produced by CParser#constantExpression.
    def exitConstantExpression(self, ctx:CParser.ConstantExpressionContext):
        pass


    # Enter a parse tree produced by CParser#declaration.
    def enterDeclaration(self, ctx:CParser.DeclarationContext):
        pass

    # Exit a parse tree produced by CParser#declaration.
    def exitDeclaration(self, ctx:CParser.DeclarationContext):
        pass


    # Enter a parse tree produced by CParser#declarationSpecifiers.
    def enterDeclarationSpecifiers(self, ctx:CParser.DeclarationSpecifiersContext):
        pass

    # Exit a parse tree produced by CParser#declarationSpecifiers.
    def exitDeclarationSpecifiers(self, ctx:CParser.DeclarationSpecifiersContext):
        pass


    # Enter a parse tree produced by CParser#declarationSpecifiers2.
    def enterDeclarationSpecifiers2(self, ctx:CParser.DeclarationSpecifiers2Context):
        pass

    # Exit a parse tree produced by CParser#declarationSpecifiers2.
    def exitDeclarationSpecifiers2(self, ctx:CParser.DeclarationSpecifiers2Context):
        pass


    # Enter a parse tree produced by CParser#declarationSpecifier.
    def enterDeclarationSpecifier(self, ctx:CParser.DeclarationSpecifierContext):
        pass

    # Exit a parse tree produced by CParser#declarationSpecifier.
    def exitDeclarationSpecifier(self, ctx:CParser.DeclarationSpecifierContext):
        pass


    # Enter a parse tree produced by CParser#initDeclaratorList.
    def enterInitDeclaratorList(self, ctx:CParser.InitDeclaratorListContext):
        pass

    # Exit a parse tree produced by CParser#initDeclaratorList.
    def exitInitDeclaratorList(self, ctx:CParser.InitDeclaratorListContext):
        pass


    # Enter a parse tree produced by CParser#initDeclarator.
    def enterInitDeclarator(self, ctx:CParser.InitDeclaratorContext):
        pass

    # Exit a parse tree produced by CParser#initDeclarator.
    def exitInitDeclarator(self, ctx:CParser.InitDeclaratorContext):
        pass


    # Enter a parse tree produced by CParser#storageClassSpecifier.
    def enterStorageClassSpecifier(self, ctx:CParser.StorageClassSpecifierContext):
        pass

    # Exit a parse tree produced by CParser#storageClassSpecifier.
    def exitStorageClassSpecifier(self, ctx:CParser.StorageClassSpecifierContext):
        pass


    # Enter a parse tree produced by CParser#typeSpecifier.
    def enterTypeSpecifier(self, ctx:CParser.TypeSpecifierContext):
        pass

    # Exit a parse tree produced by CParser#typeSpecifier.
    def exitTypeSpecifier(self, ctx:CParser.TypeSpecifierContext):
        pass


    # Enter a parse tree produced by CParser#structOrUnionSpecifier.
    def enterStructOrUnionSpecifier(self, ctx:CParser.StructOrUnionSpecifierContext):
        pass

    # Exit a parse tree produced by CParser#structOrUnionSpecifier.
    def exitStructOrUnionSpecifier(self, ctx:CParser.StructOrUnionSpecifierContext):
        pass


    # Enter a parse tree produced by CParser#structOrUnion.
    def enterStructOrUnion(self, ctx:CParser.StructOrUnionContext):
        pass

    # Exit a parse tree produced by CParser#structOrUnion.
    def exitStructOrUnion(self, ctx:CParser.StructOrUnionContext):
        pass


    # Enter a parse tree produced by CParser#structDeclarationList.
    def enterStructDeclarationList(self, ctx:CParser.StructDeclarationListContext):
        pass

    # Exit a parse tree produced by CParser#structDeclarationList.
    def exitStructDeclarationList(self, ctx:CParser.StructDeclarationListContext):
        pass


    # Enter a parse tree produced by CParser#structDeclaration.
    def enterStructDeclaration(self, ctx:CParser.StructDeclarationContext):
        pass

    # Exit a parse tree produced by CParser#structDeclaration.
    def exitStructDeclaration(self, ctx:CParser.StructDeclarationContext):
        pass


    # Enter a parse tree produced by CParser#specifierQualifierList.
    def enterSpecifierQualifierList(self, ctx:CParser.SpecifierQualifierListContext):
        pass

    # Exit a parse tree produced by CParser#specifierQualifierList.
    def exitSpecifierQualifierList(self, ctx:CParser.SpecifierQualifierListContext):
        pass


    # Enter a parse tree produced by CParser#structDeclaratorList.
    def enterStructDeclaratorList(self, ctx:CParser.StructDeclaratorListContext):
        pass

    # Exit a parse tree produced by CParser#structDeclaratorList.
    def exitStructDeclaratorList(self, ctx:CParser.StructDeclaratorListContext):
        pass


    # Enter a parse tree produced by CParser#structDeclarator.
    def enterStructDeclarator(self, ctx:CParser.StructDeclaratorContext):
        pass

    # Exit a parse tree produced by CParser#structDeclarator.
    def exitStructDeclarator(self, ctx:CParser.StructDeclaratorContext):
        pass


    # Enter a parse tree produced by CParser#enumSpecifier.
    def enterEnumSpecifier(self, ctx:CParser.EnumSpecifierContext):
        pass

    # Exit a parse tree produced by CParser#enumSpecifier.
    def exitEnumSpecifier(self, ctx:CParser.EnumSpecifierContext):
        pass


    # Enter a parse tree produced by CParser#enumeratorList.
    def enterEnumeratorList(self, ctx:CParser.EnumeratorListContext):
        pass

    # Exit a parse tree produced by CParser#enumeratorList.
    def exitEnumeratorList(self, ctx:CParser.EnumeratorListContext):
        pass


    # Enter a parse tree produced by CParser#enumerator.
    def enterEnumerator(self, ctx:CParser.EnumeratorContext):
        pass

    # Exit a parse tree produced by CParser#enumerator.
    def exitEnumerator(self, ctx:CParser.EnumeratorContext):
        pass


    # Enter a parse tree produced by CParser#enumerationConstant.
    def enterEnumerationConstant(self, ctx:CParser.EnumerationConstantContext):
        pass

    # Exit a parse tree produced by CParser#enumerationConstant.
    def exitEnumerationConstant(self, ctx:CParser.EnumerationConstantContext):
        pass


    # Enter a parse tree produced by CParser#atomicTypeSpecifier.
    def enterAtomicTypeSpecifier(self, ctx:CParser.AtomicTypeSpecifierContext):
        pass

    # Exit a parse tree produced by CParser#atomicTypeSpecifier.
    def exitAtomicTypeSpecifier(self, ctx:CParser.AtomicTypeSpecifierContext):
        pass


    # Enter a parse tree produced by CParser#typeQualifier.
    def enterTypeQualifier(self, ctx:CParser.TypeQualifierContext):
        pass

    # Exit a parse tree produced by CParser#typeQualifier.
    def exitTypeQualifier(self, ctx:CParser.TypeQualifierContext):
        pass


    # Enter a parse tree produced by CParser#functionSpecifier.
    def enterFunctionSpecifier(self, ctx:CParser.FunctionSpecifierContext):
        pass

    # Exit a parse tree produced by CParser#functionSpecifier.
    def exitFunctionSpecifier(self, ctx:CParser.FunctionSpecifierContext):
        pass


    # Enter a parse tree produced by CParser#alignmentSpecifier.
    def enterAlignmentSpecifier(self, ctx:CParser.AlignmentSpecifierContext):
        pass

    # Exit a parse tree produced by CParser#alignmentSpecifier.
    def exitAlignmentSpecifier(self, ctx:CParser.AlignmentSpecifierContext):
        pass


    # Enter a parse tree produced by CParser#declarator.
    def enterDeclarator(self, ctx:CParser.DeclaratorContext):
        pass

    # Exit a parse tree produced by CParser#declarator.
    def exitDeclarator(self, ctx:CParser.DeclaratorContext):
        pass


    # Enter a parse tree produced by CParser#directDeclarator.
    def enterDirectDeclarator(self, ctx:CParser.DirectDeclaratorContext):
        pass

    # Exit a parse tree produced by CParser#directDeclarator.
    def exitDirectDeclarator(self, ctx:CParser.DirectDeclaratorContext):
        pass


    # Enter a parse tree produced by CParser#vcSpecificModifer.
    def enterVcSpecificModifer(self, ctx:CParser.VcSpecificModiferContext):
        pass

    # Exit a parse tree produced by CParser#vcSpecificModifer.
    def exitVcSpecificModifer(self, ctx:CParser.VcSpecificModiferContext):
        pass


    # Enter a parse tree produced by CParser#gccDeclaratorExtension.
    def enterGccDeclaratorExtension(self, ctx:CParser.GccDeclaratorExtensionContext):
        pass

    # Exit a parse tree produced by CParser#gccDeclaratorExtension.
    def exitGccDeclaratorExtension(self, ctx:CParser.GccDeclaratorExtensionContext):
        pass


    # Enter a parse tree produced by CParser#gccAttributeSpecifier.
    def enterGccAttributeSpecifier(self, ctx:CParser.GccAttributeSpecifierContext):
        pass

    # Exit a parse tree produced by CParser#gccAttributeSpecifier.
    def exitGccAttributeSpecifier(self, ctx:CParser.GccAttributeSpecifierContext):
        pass


    # Enter a parse tree produced by CParser#gccAttributeList.
    def enterGccAttributeList(self, ctx:CParser.GccAttributeListContext):
        pass

    # Exit a parse tree produced by CParser#gccAttributeList.
    def exitGccAttributeList(self, ctx:CParser.GccAttributeListContext):
        pass


    # Enter a parse tree produced by CParser#gccAttribute.
    def enterGccAttribute(self, ctx:CParser.GccAttributeContext):
        pass

    # Exit a parse tree produced by CParser#gccAttribute.
    def exitGccAttribute(self, ctx:CParser.GccAttributeContext):
        pass


    # Enter a parse tree produced by CParser#nestedParenthesesBlock.
    def enterNestedParenthesesBlock(self, ctx:CParser.NestedParenthesesBlockContext):
        pass

    # Exit a parse tree produced by CParser#nestedParenthesesBlock.
    def exitNestedParenthesesBlock(self, ctx:CParser.NestedParenthesesBlockContext):
        pass


    # Enter a parse tree produced by CParser#pointer.
    def enterPointer(self, ctx:CParser.PointerContext):
        pass

    # Exit a parse tree produced by CParser#pointer.
    def exitPointer(self, ctx:CParser.PointerContext):
        pass


    # Enter a parse tree produced by CParser#typeQualifierList.
    def enterTypeQualifierList(self, ctx:CParser.TypeQualifierListContext):
        pass

    # Exit a parse tree produced by CParser#typeQualifierList.
    def exitTypeQualifierList(self, ctx:CParser.TypeQualifierListContext):
        pass


    # Enter a parse tree produced by CParser#parameterTypeList.
    def enterParameterTypeList(self, ctx:CParser.ParameterTypeListContext):
        pass

    # Exit a parse tree produced by CParser#parameterTypeList.
    def exitParameterTypeList(self, ctx:CParser.ParameterTypeListContext):
        pass


    # Enter a parse tree produced by CParser#parameterList.
    def enterParameterList(self, ctx:CParser.ParameterListContext):
        pass

    # Exit a parse tree produced by CParser#parameterList.
    def exitParameterList(self, ctx:CParser.ParameterListContext):
        pass


    # Enter a parse tree produced by CParser#parameterDeclaration.
    def enterParameterDeclaration(self, ctx:CParser.ParameterDeclarationContext):
        pass

    # Exit a parse tree produced by CParser#parameterDeclaration.
    def exitParameterDeclaration(self, ctx:CParser.ParameterDeclarationContext):
        pass


    # Enter a parse tree produced by CParser#identifierList.
    def enterIdentifierList(self, ctx:CParser.IdentifierListContext):
        pass

    # Exit a parse tree produced by CParser#identifierList.
    def exitIdentifierList(self, ctx:CParser.IdentifierListContext):
        pass


    # Enter a parse tree produced by CParser#typeName.
    def enterTypeName(self, ctx:CParser.TypeNameContext):
        pass

    # Exit a parse tree produced by CParser#typeName.
    def exitTypeName(self, ctx:CParser.TypeNameContext):
        pass


    # Enter a parse tree produced by CParser#abstractDeclarator.
    def enterAbstractDeclarator(self, ctx:CParser.AbstractDeclaratorContext):
        pass

    # Exit a parse tree produced by CParser#abstractDeclarator.
    def exitAbstractDeclarator(self, ctx:CParser.AbstractDeclaratorContext):
        pass


    # Enter a parse tree produced by CParser#directAbstractDeclarator.
    def enterDirectAbstractDeclarator(self, ctx:CParser.DirectAbstractDeclaratorContext):
        pass

    # Exit a parse tree produced by CParser#directAbstractDeclarator.
    def exitDirectAbstractDeclarator(self, ctx:CParser.DirectAbstractDeclaratorContext):
        pass


    # Enter a parse tree produced by CParser#typedefName.
    def enterTypedefName(self, ctx:CParser.TypedefNameContext):
        pass

    # Exit a parse tree produced by CParser#typedefName.
    def exitTypedefName(self, ctx:CParser.TypedefNameContext):
        pass


    # Enter a parse tree produced by CParser#initializer.
    def enterInitializer(self, ctx:CParser.InitializerContext):
        pass

    # Exit a parse tree produced by CParser#initializer.
    def exitInitializer(self, ctx:CParser.InitializerContext):
        pass


    # Enter a parse tree produced by CParser#initializerList.
    def enterInitializerList(self, ctx:CParser.InitializerListContext):
        pass

    # Exit a parse tree produced by CParser#initializerList.
    def exitInitializerList(self, ctx:CParser.InitializerListContext):
        pass


    # Enter a parse tree produced by CParser#designation.
    def enterDesignation(self, ctx:CParser.DesignationContext):
        pass

    # Exit a parse tree produced by CParser#designation.
    def exitDesignation(self, ctx:CParser.DesignationContext):
        pass


    # Enter a parse tree produced by CParser#designatorList.
    def enterDesignatorList(self, ctx:CParser.DesignatorListContext):
        pass

    # Exit a parse tree produced by CParser#designatorList.
    def exitDesignatorList(self, ctx:CParser.DesignatorListContext):
        pass


    # Enter a parse tree produced by CParser#designator.
    def enterDesignator(self, ctx:CParser.DesignatorContext):
        pass

    # Exit a parse tree produced by CParser#designator.
    def exitDesignator(self, ctx:CParser.DesignatorContext):
        pass


    # Enter a parse tree produced by CParser#staticAssertDeclaration.
    def enterStaticAssertDeclaration(self, ctx:CParser.StaticAssertDeclarationContext):
        pass

    # Exit a parse tree produced by CParser#staticAssertDeclaration.
    def exitStaticAssertDeclaration(self, ctx:CParser.StaticAssertDeclarationContext):
        pass


    # Enter a parse tree produced by CParser#statement.
    def enterStatement(self, ctx:CParser.StatementContext):
        pass

    # Exit a parse tree produced by CParser#statement.
    def exitStatement(self, ctx:CParser.StatementContext):
        pass


    # Enter a parse tree produced by CParser#labeledStatement.
    def enterLabeledStatement(self, ctx:CParser.LabeledStatementContext):
        pass

    # Exit a parse tree produced by CParser#labeledStatement.
    def exitLabeledStatement(self, ctx:CParser.LabeledStatementContext):
        pass


    # Enter a parse tree produced by CParser#compoundStatement.
    def enterCompoundStatement(self, ctx:CParser.CompoundStatementContext):
        pass

    # Exit a parse tree produced by CParser#compoundStatement.
    def exitCompoundStatement(self, ctx:CParser.CompoundStatementContext):
        pass


    # Enter a parse tree produced by CParser#blockItemList.
    def enterBlockItemList(self, ctx:CParser.BlockItemListContext):
        pass

    # Exit a parse tree produced by CParser#blockItemList.
    def exitBlockItemList(self, ctx:CParser.BlockItemListContext):
        pass


    # Enter a parse tree produced by CParser#blockItem.
    def enterBlockItem(self, ctx:CParser.BlockItemContext):
        pass

    # Exit a parse tree produced by CParser#blockItem.
    def exitBlockItem(self, ctx:CParser.BlockItemContext):
        pass


    # Enter a parse tree produced by CParser#expressionStatement.
    def enterExpressionStatement(self, ctx:CParser.ExpressionStatementContext):
        pass

    # Exit a parse tree produced by CParser#expressionStatement.
    def exitExpressionStatement(self, ctx:CParser.ExpressionStatementContext):
        pass


    # Enter a parse tree produced by CParser#selectionStatement.
    def enterSelectionStatement(self, ctx:CParser.SelectionStatementContext):
        pass

    # Exit a parse tree produced by CParser#selectionStatement.
    def exitSelectionStatement(self, ctx:CParser.SelectionStatementContext):
        pass


    # Enter a parse tree produced by CParser#iterationStatement.
    def enterIterationStatement(self, ctx:CParser.IterationStatementContext):
        pass

    # Exit a parse tree produced by CParser#iterationStatement.
    def exitIterationStatement(self, ctx:CParser.IterationStatementContext):
        pass


    # Enter a parse tree produced by CParser#multiLineMacroStatement.
    def enterMultiLineMacroStatement(self, ctx:CParser.MultiLineMacroStatementContext):
        pass

    # Exit a parse tree produced by CParser#multiLineMacroStatement.
    def exitMultiLineMacroStatement(self, ctx:CParser.MultiLineMacroStatementContext):
        pass


    # Enter a parse tree produced by CParser#forCondition.
    def enterForCondition(self, ctx:CParser.ForConditionContext):
        pass

    # Exit a parse tree produced by CParser#forCondition.
    def exitForCondition(self, ctx:CParser.ForConditionContext):
        pass


    # Enter a parse tree produced by CParser#forDeclaration.
    def enterForDeclaration(self, ctx:CParser.ForDeclarationContext):
        pass

    # Exit a parse tree produced by CParser#forDeclaration.
    def exitForDeclaration(self, ctx:CParser.ForDeclarationContext):
        pass


    # Enter a parse tree produced by CParser#forExpression.
    def enterForExpression(self, ctx:CParser.ForExpressionContext):
        pass

    # Exit a parse tree produced by CParser#forExpression.
    def exitForExpression(self, ctx:CParser.ForExpressionContext):
        pass


    # Enter a parse tree produced by CParser#jumpStatement.
    def enterJumpStatement(self, ctx:CParser.JumpStatementContext):
        pass

    # Exit a parse tree produced by CParser#jumpStatement.
    def exitJumpStatement(self, ctx:CParser.JumpStatementContext):
        pass


    # Enter a parse tree produced by CParser#compilationUnit.
    def enterCompilationUnit(self, ctx:CParser.CompilationUnitContext):
        pass

    # Exit a parse tree produced by CParser#compilationUnit.
    def exitCompilationUnit(self, ctx:CParser.CompilationUnitContext):
        pass


    # Enter a parse tree produced by CParser#translationUnit.
    def enterTranslationUnit(self, ctx:CParser.TranslationUnitContext):
        pass

    # Exit a parse tree produced by CParser#translationUnit.
    def exitTranslationUnit(self, ctx:CParser.TranslationUnitContext):
        pass


    # Enter a parse tree produced by CParser#externalDeclaration.
    def enterExternalDeclaration(self, ctx:CParser.ExternalDeclarationContext):
        pass

    # Exit a parse tree produced by CParser#externalDeclaration.
    def exitExternalDeclaration(self, ctx:CParser.ExternalDeclarationContext):
        pass


    # Enter a parse tree produced by CParser#functionDefinition.
    def enterFunctionDefinition(self, ctx:CParser.FunctionDefinitionContext):
        pass

    # Exit a parse tree produced by CParser#functionDefinition.
    def exitFunctionDefinition(self, ctx:CParser.FunctionDefinitionContext):
        pass


    # Enter a parse tree produced by CParser#declarationList.
    def enterDeclarationList(self, ctx:CParser.DeclarationListContext):
        pass

    # Exit a parse tree produced by CParser#declarationList.
    def exitDeclarationList(self, ctx:CParser.DeclarationListContext):
        pass



del CParser


import unittest
from mccode_antlr.assembler import Assembler


def curved_guide_inserter():
    from numpy import ones, hstack
    small_distance = 1e-05  # m, Small value to separate components
    curved_guide_lengths = 0.5 * ones(3)  # 36 segments approximating the curve, but two are broken into two guides
    curved_guide_angles = 0.01886551 * ones(3)  # this is 'benderAngle' from the original file
    curved_guide_x_translations = 7.57345916228e-10 * ones(3)
    curved_guide_x_translations[0] = 0.
    # Each guide segment was either 'u', 'u + 6e-5 + [previous length]', or '0 + [previous length]' translated along z
    curved_guide_z_translations = hstack((small_distance, 7 * small_distance + curved_guide_lengths[:-1]))

    def add_curved_guide_piece(assembler, number):
        # guide pieces are translated along beam path, and moved sideways due to curvature
        at = curved_guide_x_translations[number], 0, curved_guide_z_translations[number]
        # rotated around the vertical axis
        rot = 0, curved_guide_angles[number], 0
        rel = 'end_element6' if number < 1 else f'curved_guide_segment_{number - 1:02d}'
        seg_name = f'curved_guide_segment_{number:02d}'
        seg = assembler.component(seg_name, 'Guide_gravity', at=(at, rel), rotate=(rot, rel))
        seg.set_parameters(w1=0.029534, h1=0.047514, w2=0.029534, h2=0.047514, l=curved_guide_lengths[number], R0=0.99,
                           Qc=0.0217, alpha=3.1, W=0.003, mleft=3.0, mright=3.5, mtop=2.5, mbottom=2.5, G=-9.82)
        end_name = f'{seg_name}_end'
        assembler.component(end_name, 'Arm', at=((0, 0, curved_guide_lengths[number] + 1e-6), seg_name))
        return seg_name, end_name

    return add_curved_guide_piece


def add_elliptic_guide(assembler, name, at=None, relative=None, **kwargs):
    if at is None:
        at = [0, 0, 1e-5]
    guide = assembler.component(name, 'Elliptic_guide_gravity', at=(at, relative))
    guide.set_parameters(dimensionsAt="\"mid\"", R0=0.99, Qc=0.0217, alpha=3.1, W=0.003, **kwargs)
    return guide


def declare_nboa_guide_parameters(assembler):
    def declare_line(name, vec):
        initializer = '{' + ','.join(f'{x}' for x in vec) + '}'
        return f'double {name}[] = {initializer};'

    # Neutron Beam Optical Assembly (NBOA) in-monolith guide:
    nboa = dict()
    nboa['nboa_lens'] = [
        0.48844444, 0.48844444, 0.48844444, 0.48844444, 0.48844444, 0.48844444, 0.48844444, 0.06121889, 0.02349,
        0.40373556, 0.07626444, 0.015, 0.39671
    ]
    nboa['nboa_vert'] = [3.5, 3.5, 3.0, 3.0, 2.5, 2.5, 2.5, 2.5, 0.0, 2.0, 2.0, 0.0, 2.0]
    nboa['nboa_horz'] = [3.0, 3.0, 2.5, 2.5, 2.5, 2.0, 1.5, 1.5, 0.0, 1.5, 1.5, 0.0, 1.5]

    assembler.declare('\n'.join([declare_line(n, v) for n, v in nboa.items()]), source=__file__)

    return nboa['nboa_lens']


def insert_truncated_bifrost_primary(b: Assembler):
    import numpy as np

    eps = 1e-05  # m, Small value to separate components

    el6_lens = declare_nboa_guide_parameters(b)

    # call-time defined instrument parameters
    parameters = [
        'int pulses = 1', 'power/"MW" = 2',
        'source_lambda_min/"angstrom" = 0.75', 'source_lambda_max/"angstrom" = 30',
    ]
    for parameter in parameters:
        b.parameter(parameter)

    # Start the actual instrument:
    b.component("Origin", "Progress_bar", at=[0, 0, 0])

    # Element 6 (NBOA) start position relative the source
    el6_s = np.array([0.01277, 0, 1.903398 - eps])
    # Element 6 (NBOA) end position relative its start -- it's length
    el6_d = np.array([0, 0, 4.39553 + 2 * eps])

    source_to_nboa_entrance = 1.903398
    nboa_entrance_width = 0.068797 + 2 * 0.01277
    nboa_entrance_height = 0.03472
    curved_guide_reflections_count = None
    b.component("ESS_source", "ESS_butterfly", at=[0, 0, 0], rotate=[0, 0, 0]) \
        .set_parameters(sector="\"W\"", beamline=4, yheight=0.03, cold_frac=0.5, dist=source_to_nboa_entrance,
                        focus_xw=nboa_entrance_width, focus_yh=nboa_entrance_height,
                        c_performance=1.0, t_performance=1.0, Lmin="source_lambda_min", Lmax="source_lambda_max",
                        tmax_multiplier=1.500000, n_pulses="pulses", acc_power="power",
                        # tfocus_dist="tf_d", tfocus_time="tf_t", tfocus_width="1.1*tf_w"
                        )

    b.component("GuideStart", "Arm", at=(list(el6_s), 'Origin'), rotate=([0, -0.56, 0], 'Origin'))

    # The NBOA is element6 -- its return value is not used and originally had the parameter l specified twice: l=4.39553
    add_elliptic_guide(b, "NBOA", relative='GuideStart',
                       xwidth=0.069634, yheight=0.04862, linxw=3.4578, loutxw=0.415155, linyh=1.36, loutyh=3.487681,
                       seglength="nboa_lens", mvaluesright="nboa_horz", mvaluesleft="nboa_horz",
                       mvaluestop="nboa_vert", mvaluesbottom="nboa_vert", nSegments=len(el6_lens),
                       # l="+".join([f"{x}" for x in el6_lens])  # Push the sum to C, to avoid rounding errors?
                       l=sum(el6_lens)
                       )

    b.component("end_element6", "Arm", at=(list(el6_d), "PREVIOUS"))
    b.component("end_element5", "Arm", at=([0, 0, 0.081707], "end_element6"))

    add_curved_guide_piece = curved_guide_inserter()
    guide_0_segment, guide_0_end = add_curved_guide_piece(b, 0)
    guide_1_segment, guide_1_end = add_curved_guide_piece(b, 1)

    split_at = b.component('split_at', 'Arm', at=([0, 0, 2 * eps], guide_1_end))

    sample_stack = b.component("sample_stack", "Arm", at=([0, 0, 0.58], split_at))

    return b, split_at, sample_stack


def make_truncated_bifrost():
    from mccode_antlr.assembler import Assembler
    from mccode_antlr.reader import MCSTAS_REGISTRY
    bifrost = Assembler('BIFROST', registries=[MCSTAS_REGISTRY])
    bifrost, split_at, sample_stack = insert_truncated_bifrost_primary(bifrost)
    return bifrost.instrument


class SplitBifrostTestCase(unittest.TestCase):
    def setUp(self):
        self.bifrost = make_truncated_bifrost()

    def assertBothReferencesEqual(self, comp, ref):
        self.assertEqual(comp.at_relative[1], ref)
        self.assertEqual(comp.rotate_relative[1], ref)

    def _test_relative_references(self, instr):
        origin = instr.get_component('Origin')
        guide_start = instr.get_component('GuideStart')
        nboa = instr.get_component('NBOA')
        end_element_6 = instr.get_component('end_element6')
        guide0 = instr.get_component('curved_guide_segment_00')
        guide1 = instr.get_component('curved_guide_segment_01')
        self.assertBothReferencesEqual(guide_start, origin)
        self.assertBothReferencesEqual(nboa, guide_start)
        self.assertBothReferencesEqual(end_element_6, nboa)
        self.assertBothReferencesEqual(guide0, end_element_6)
        self.assertBothReferencesEqual(guide1, guide0)

    def test_pre(self):
        self._test_relative_references(self.bifrost)

    def test_split(self):
        before, after = self.bifrost.mcpl_split(self.bifrost.get_component('split_at'))
        self._test_relative_references(before)


class SplitHDF5BifrostTestCase(SplitBifrostTestCase):
    def setUp(self):
        from pathlib import Path
        from tempfile import mkdtemp
        from importlib.util import find_spec
        if not find_spec('h5py'):
            self.skipTest('h5py not found -- skipping HDF5 tests')
        else:
            from mccode_antlr.io import save_hdf5, load_hdf5
            self.td = Path(mkdtemp())
            filepath = self.td.joinpath('bifrost.h5')
            save_hdf5(make_truncated_bifrost(), filepath)
            self.bifrost = load_hdf5(filepath)

    def tearDown(self):
        from shutil import rmtree
        rmtree(self.td)


if __name__ == '__main__':
    unittest.main()



"""Monte Carlo Particle Ray Tracing compiler, Volume 4"""
import mccode_antlr.grammar
__author__ = "Gregory Tucker"
__affiliation__ = "European Spallation Source ERIC"

from .version import version
__version__ = version()

__all__ = ["__author__", "__affiliation__", "__version__", "version"]



from .c_listener import CDeclarator
from ..comp import Comp
from ..instr import Instr

def _runtime_parameters(is_mcstas: bool):
    pars = ['x', 'y', 'z']
    if is_mcstas:
        pars.extend(['vx', 'vy', 'vz', 't', 'sx', 'sy', 'sz', 'p', 'mcgravitation', 'mcMagnet', 'allow_backprop'])
    else:
        pars.extend(['kx', 'ky', 'kz', 'phi', 't', 'Ex', 'Ey', 'Ez', 'p'])
    # Add the temporary particle parameters added sometime after v3.4.0 ... to the v3.4.0 tag
    pars.extend(['_mctmp_a', '_mctmp_b', '_mctmp_c'])
    return pars


def _runtime_kv_parameters(is_mcstas: bool):
    pars = ['p', 't']
    pars.extend(['vx', 'vy', 'vz'] if is_mcstas else ['kx', 'ky', 'kz'])
    pars.extend(['x', 'y', 'z'])
    return pars


def def_trace_section(is_mcstas: bool):
    lines = [
        "/*******************************************************************************",
        "* components TRACE",
        "*******************************************************************************/"
    ]
    lines.extend([f'#define {x} (_particle->{x})' for x in _runtime_parameters(is_mcstas)])
    lines.extend([
        "/* if on GPU, globally nullify sprintf,fprintf,printfs   */",
        "/* (Similar defines are available in each comp trace but */",
        "/*  those are not enough to handle external libs etc. )  */",
        "#ifdef OPENACC",
        "#define fprintf(stderr,...) printf(__VA_ARGS__)",
        "#define sprintf(string,...) printf(__VA_ARGS__)",
        "#define exit(...) noprintf()",
        "#define strcmp(a,b) str_comp(a,b)",
        "#define strlen(a) str_len(a)",
        "#endif",
        # /* define SCATTERED, ABSORB and RESTORE macros for all TRACE */"
        "#define SCATTERED (_particle->_scattered)",
        "#define RESTORE (_particle->_restore)",
        "#define ABSORBED (_particle->_absorbed)",
        "#define RESTORE_NEUTRON(_index, ...) _particle->_restore = _index;",
        # /* define mcget_run_num within trace scope to refer to the particle */
        "#define mcget_run_num() _particle->_uid",
        "#define ABSORB0 do { DEBUG_STATE(); DEBUG_ABSORB(); MAGNET_OFF; ABSORBED++; return(_comp); } while(0)",
        "#define ABSORB ABSORB0"
    ])
    return '\n'.join(lines)


def undef_trace_section(is_mcstas: bool):
    lines = [f'#undef {x}' for x in _runtime_parameters(is_mcstas)]
    lines.extend([
        "#ifdef OPENACC",
        "#ifndef MULTICORE",
        "#undef strlen",
        "#undef strcmp",
        "#undef exit",
        "#undef printf",
        "#undef sprintf",
        "#undef fprintf",
        "#endif",
        "#endif",
        "#undef SCATTERED",
        "#undef RESTORE",
        "#undef RESTORE_NEUTRON",
        "#undef STORE_NEUTRON",
        "#undef ABSORBED",
        "#undef ABSORB",
        "#undef ABSORB0",
    ])
    return '\n'.join(lines)


def cogen_trace_section(
        is_mcstas: bool,
        source: Instr,
        declared_parameters: dict[str, list[CDeclarator]],
        instrument_uservars: list[CDeclarator],
        component_uservars: dict[str, list[CDeclarator]],
) -> str:
    return '\n'.join([
        cogen_comp_trace_class(
            is_mcstas,
            component_type,
            source,
            declared_parameters[component_type.name],
            instrument_uservars,
            component_uservars[component_type.name]
        ) for component_type in source.component_types()
    ])


def cogen_comp_trace_class(
        is_mcstas: bool,
        comp: Comp,
        source: Instr,
        declared_parameters: list[CDeclarator],
        instr_uservars: list[CDeclarator],
        comp_uservars: list[CDeclarator],
) -> str:
    from .c_defines import cogen_parameter_define, cogen_parameter_undef
    # count matching component type instances which define an EXTEND block:
    extended = [(n, i) for n, i in enumerate(source.components) if i.type.name == comp.name and len(i.extend)]

    if len(extended) == 0 and all(x.is_empty for x in comp.trace):
        return ''

    lines = [
        '#pragma acc routine',
        f'_class_{comp.name} *class_{comp.name}_trace(_class_{comp.name} *_comp, _class_particle *_particle) {{',
        'ABSORBED=SCATTERED=RESTORE=0;',
        cogen_parameter_define(comp, declared_parameters),
    ]
    f, n = comp.trace[0].fn if len(comp.trace) else (comp.name, 0)
    lines.append(f'SIG_MESSAGE("[_{comp.name}_trace] component NULL={comp.name}() [{f}:{n}]");')

    # FIXME This should be checking *only* for file.comp *formal* parameters specified with `symbol` as the typename
    #       which is not yet supported in this translator. Only the formal parameters can have `instr_type_symbol`
    #       and only in two specific situations: symbol {name}, symbol {name} = {expr}

    # Check if there are any user-defined parameter types ... (something which wasn't set previously?)
    # This is the 'symbol' type
    declared_types = [x.dtype for x in declared_parameters]
    # there must be a better way than this
    is_symbol = [t == 'symbol' for t in declared_types]
    # TODO FIXME This should be looping through setting parameters. It is probably wrong.
    if any(is_symbol):
        for i, inst in [(i, inst) for i, inst in enumerate(source.components) if inst.type.name == comp.name]:
            lines.extend([
                f'  /* check if this is component {inst.name} */',
                f'  if(_comp->_index == {1+i}) {{ '
            ])
            # loop through the symbol types and set the component-instance values from ... somewhere
            for c_dec in declared_parameters:
                # Use the user-defined instance parameter if it exists, or attempt to use a default otherwise?
                v = next(iter(p for p in inst.parameters if p.name == c_dec.name), None)
                v = v.value if v else c_dec.init
                if v is not None:
                    lines.append(f'    {c_dec.name} = {v};')
            lines.append('  }')

    # output the actual TRACE block(s)
    for block in comp.trace:
        lines.append(block.to_c())

    # instr files do not produce a code block to output here.
    pars = _runtime_kv_parameters(is_mcstas)
    lines.extend(["#ifndef NOABSORB_INF_NAN", "  /* Check for nan or inf particle parms */ "])
    # Change following https://github.com/McStasMcXtrace/McCode/commit/6876e5c6f6618d527c67a01610119fa8b1188084
    # lines.extend([f"  if(isnan({x}) || isinf({x})) ABSORB;" for x in pars])
    lines.append(f"  if(isnan({' + '.join(pars)})) ABSORB;")
    lines.append(f"  if(isinf({' + '.join(f'''fabs({x})''' for x in pars)})) ABSORB;")
    lines.append("#else")

    def long(x):
        return f'printf("NAN or INF found in {x}, %s (particle %lld)\\n",_comp->_name,_particle->_uid)'

    lines.extend([f'  if(isnan({x}) || isinf({x})) {long(x)};' for x in pars])
    lines.append('#endif')

    if len(extended):
        # combine the USERVARS from the instrument and this component type blocks:
        uvs = list(dict.fromkeys([*instr_uservars, *comp_uservars]))
        # So that the EXTEND block(s) can access them
        lines.extend([f'  #define {x.name} (_particle->{x.name})' for x in uvs])
        # `index` was defined above to be the component index into the full instrument list
        for index, inst in extended:
            lines.append(f'if (_comp->_index == {1+index}) {{ // EXTEND {inst.name}')
            for ext in inst.extend:
                lines.append(ext.to_c())
            lines.append("}")
        lines.extend(f'  #undef {x.name}' for x in uvs)

    # return the component
    lines.extend([
        cogen_parameter_undef(comp, declared_parameters),
        '  return(_comp);',
        f'}} /* class_{comp.name}_trace */',
        ''
    ])

    return '\n'.join(lines)



from unittest import TestCase
from loguru import logger
from .utils import make_assembler



class TestInstrInstanceParameters(TestCase):
    def test_assemble_identifier_instance_parameter(self):
        assembler = make_assembler('fake_bifrost')

        parameters = [
            'ps1speed/"Hz" = 196', 'ps2speed/"Hz" = 196',
            'fo1speed/"Hz" = 14', 'fo2speed/"Hz" = 14',
            'bw1speed/"Hz" = 14', 'bw2speed/"Hz" = -14',
            'ps1phase/"Hz" = 0', 'ps2phase/"Hz" = 0',
            'fo1phase/"Hz" = 0', 'fo2phase/"Hz" = 0',
            'bw1phase/"Hz" = 0', 'bw2phase/"Hz" = 0',
            'blade0l/"m" = -0.05', 'blade0r/"m" = 0.05',
            'blade1l/"m" = -0.05', 'blade1r/"m" = 0.05',
            'blade2l/"m" = -0.05', 'blade2r/"m" = 0.05',
            'blade3l/"m" = -0.05', 'blade3r/"m" = 0.05',
            'int pulses = 1', 'power/"MW" = 2',
            'source_lambda_min/"angstrom" = 0.75', 'source_lambda_max/"angstrom" = 30',
        ]
        assembler.parameters(*parameters)

        # Mix of initialization temporary variables and calculated instrument parameters.
        pulse_length = 2.86E-3
        assembler.declare(f"""
         double lambda_0;
         double lambda_1;
         double tf_d=1;
         double tf_t={pulse_length / 2};
         double tf_w={pulse_length};
         double tf_tau={1 / 14.0};
         double ps_disk_open_angle=170.0;
         double ps_distance={4.41 + 0.032 + 2.0 - 0.1};
         double bw_disk_open_angle=161.0;
         double bw_distance=78.0;
         """)

        # Chopper-based wavelength calculations can only be done at runtime since they depend on instrument parameters
        new_chopper_calculations = """
         chopper_parameters * pars = calloc(6, sizeof(chopper_parameters));
         pars[0].speed = ps1speed; // Pulse shaping disk 1
         pars[0].phase = ps1phase;
         pars[0].angle = 170.0;
         pars[0].path = ps_distance;
         pars[1].speed = ps2speed; // Pulse shaping disk 2
         pars[1].phase = ps2phase;
         pars[1].angle = pars[0].angle;
         pars[1].path = pars[0].path + 0.02;
         pars[2].speed = fo1speed; // Frame overlap 1
         pars[2].phase = fo1phase;
         pars[2].angle = 38.26;
         pars[2].path = 8.530;
         pars[3].speed = fo2speed; // Frame overlap 2
         pars[3].phase = fo2phase;
         pars[3].angle = 52.01;
         pars[3].path = 14.973;
         pars[4].speed = bw1speed; // Bandwidth disk 1
         pars[4].phase = bw1phase;
         pars[4].angle = 161.0;
         pars[4].path = bw_distance;
         pars[5].speed = bw2speed; // Bandwidth disk 2
         pars[5].phase = bw2phase;
         pars[5].angle = pars[4].angle;
         pars[5].path = pars[4].path + 0.02;

         lambda_0 = source_lambda_min;
         lambda_1 = source_lambda_max;
         double pulse_delay = 2.0e-4; // time to peak brightness after proton pulse T0
         double pulse_width = 2.86e-3; // duration of high flux plateau
         double latest = pulse_delay + pulse_width + 2e-3; // extra for good measure?
         unsigned windows = chopper_wavelength_limits(&source_lambda_min, &source_lambda_max, 6, pars, lambda_0, lambda_1, latest);

         // free the memory used by pars
         if (pars) free(pars);

         if (windows == 0) {
           printf("Chopper train will not pass wavelengths between %f and %f angstrom!\\n", lambda_0, lambda_1);
           printf("Examine the provided chopper speeds and phases.\\n");
         }
         if (windows > 1) {
           printf("Chopper train will pass %u wavelength ranges between %f and %f angstrom\\n", windows, lambda_0, lambda_1);
           printf("but only their envelope between %f and %f angstrom is considered!\\n", source_lambda_min, source_lambda_max);   
         }
         printf("Using source lambda limits %f to %f\\n", source_lambda_min, source_lambda_max);

         // Now calculate the time-focusing window on the first chopper given the wavelength band:
         // ...

         """
        assembler.initialize(new_chopper_calculations)

        # Start the actual instrument:
        assembler.component("Origin", "Progress_bar", at=[0, 0, 0])

        # Elliptic guide gravity initialization and declaration
        def declare_line(name, vec):
            initializer = '{' + ','.join(f'{x}' for x in vec) + '}'
            return f'double {name}[] = {initializer};'

        # Neutron Beam Optical Assembly (NBOA) in-monolith guide:
        nboa = dict()
        nboa['nboa_lens'] = [
            0.48844444, 0.48844444, 0.48844444, 0.48844444, 0.48844444, 0.48844444, 0.48844444, 0.06121889, 0.02349,
            0.40373556, 0.07626444, 0.015, 0.39671
        ]
        nboa['nboa_vert'] = [3.5, 3.5, 3.0, 3.0, 2.5, 2.5, 2.5, 2.5, 0.0, 2.0, 2.0, 0.0, 2.0]
        nboa['nboa_horz'] = [3.0, 3.0, 2.5, 2.5, 2.5, 2.0, 1.5, 1.5, 0.0, 1.5, 1.5, 0.0, 1.5]

        assembler.declare('\n'.join([declare_line(n, v) for n, v in nboa.items()]), source=__file__)

        el6_lens = nboa['nboa_lens']

        guide = assembler.component("NBOA", 'Elliptic_guide_gravity', at=([0, 0, 1e-5], "PREVIOUS"))
        # make sure that array-valued parameters can be set:
        guide.set_parameters(dimensionsAt="\"mid\"", R0=0.99, Qc=0.0217, alpha=3.1, W=0.003, xwidth=0.069634,
                             yheight=0.04862, linxw=3.4578, loutxw=0.415155, linyh=1.36, loutyh=3.487681,
                             seglength="nboa_lens", mvaluesright="nboa_horz", mvaluesleft="nboa_horz",
                             mvaluestop="nboa_vert", mvaluesbottom="nboa_vert", nSegments=len(el6_lens),
                             l=sum(el6_lens))

        # Bandwidth chopper
        bw_upstream_gap = 0.1
        bwc1 = assembler.component("BWC1", "DiskChopper", at=([0, 0, bw_upstream_gap], "PREVIOUS"))
        bwc1.set_parameters(theta_0="bw_disk_open_angle", radius='1+2+3+4', nslit='(2+4)/(1+2)', yheight=0.1,
                            nu="bw1speed", phase="bw1phase/2")

        for par in ('theta_0', 'nslit', 'nu', 'phase'):
            self.assertFalse(bwc1.get_parameter(par).value.is_constant)
        for par in ('radius', 'yheight'):
            self.assertTrue(bwc1.get_parameter(par).value.is_constant)

        # nslit is not a constant parameter because the expression parsing does not simplify grouped expressions
        nslit = bwc1.get_parameter('nslit').value.simplify()
        # But explicitly simplifying the expression is possible (in this simple case) and yields a constant value
        self.assertTrue(nslit.simplify().is_constant)

        phase = bwc1.get_parameter('phase').value  # == bw1phase / 2
        # In contrast, phase is Expr(BinaryOp('/', Expr.id('bw1phase'), Expr.int(2))) which can't be simplified
        self.assertFalse(phase.simplify().is_constant)

        # But we can attempt to parse the declarations and instantiation blocks from the instrument
        from mccode_antlr.translators.c_listener import extract_c_declared_expressions, evaluate_c_defined_expressions
        # Here variables _is_ and _must be_ dict[str, Expr]
        variables = {x.name: x.value for x in assembler.instrument.parameters}
        for dec in assembler.instrument.declare:
            # This produces dict[CDeclarator, Expr]
            decs_expr = extract_c_declared_expressions(dec.source)
            # So extract just the CDeclarator's name
            variables.update({key.name: val for key, val in decs_expr.items()})

        # defined as
        # TODO this does not work because the simple "C"-style expression parser doesn't know about pointers
        try:
            for init in assembler.instrument.initialize:
                variables = evaluate_c_defined_expressions(variables, init.source)
        except AttributeError:
            logger.warning('Evaluating INITIALIZE section failed; see preceding errors for hints for why. '
                           'This is not a test failure condition (for now). Continuing')

        evaluated_phase = phase.evaluate(variables)
        self.assertTrue(evaluated_phase.is_constant)

        theta_0 = bwc1.get_parameter('theta_0').value
        self.assertTrue(theta_0.evaluate(variables).is_constant)

        # We want to know whether an expression involves any instrument parameters
        # -- which phase and nu do, but others do not:
        for x in ('phase', 'nu'):
            par = bwc1.get_parameter(x).value
            dep = any(par.depends_on(inst_par.name) for inst_par in assembler.instrument.parameters)
            self.assertTrue(dep)
        for x in ('theta_0', 'radius', 'nslit', 'yheight'):
            par = bwc1.get_parameter(x).value
            self.assertFalse(any(par.depends_on(inst_par.name) for inst_par in assembler.instrument.parameters))



from __future__ import annotations
from functools import cache

def subprocess_fails(args: list[str]):
    import subprocess
    try:
        subprocess.run(args, check=True)
        return False
    except FileNotFoundError:
        pass
    except RuntimeError:
        pass
    return True


@cache
def check_for_mccode_antlr_compiler(which: str) -> bool:
    from loguru import logger
    from ..config import config
    cc = config
    for key in which.split('/'):
        cc = cc[key]
    cc = cc.get(str)

    # different compilers support different 'version' or 'help' command line options
    options = '--version', '/?', '--help'
    if all(subprocess_fails([cc, opt]) for opt in options):
        logger.info(f"Compiler '{cc}' not found.")
        which = which.replace('/','_')
        logger.info(f'Provide alternate via MCCODE_ANTLR_{which} environment variable')
        return False
    return True


def compiles(compiler: str, instr):
    from os import access, R_OK
    from loguru import logger
    from pathlib import Path
    from tempfile import TemporaryDirectory
    from subprocess import run
    from mccode_antlr.translators.target import MCSTAS_GENERATOR
    from .c import CBinaryTarget, get_compiler_linker_flags, instrument_source
    from ..config import config as module_config

    target = CBinaryTarget(mpi='mpi' in compiler, acc=compiler == 'acc', count=1, nexus=False)

    compile_config = dict(default_main=True, enable_trace=False, portable=False,
                  include_runtime=True, embed_instrument_file=False, verbose=False)

    compiler_flags, linker_flags = get_compiler_linker_flags(instr, target)

    with TemporaryDirectory() as directory:
        binary = Path(directory) / f"output{module_config['ext'].get(str)}"
        command = [target.compiler, *compiler_flags, '-o', str(binary), '-', *linker_flags]
        source = instrument_source(instr, generator=MCSTAS_GENERATOR, config=compile_config)
        result = run(command, input=source, text=True, capture_output=True)

        if result.returncode:
            logger.info(f'Failed compiling simple instrument with error: {result.stderr}')
            logger.info(f"Verify compiler {target.compiler} accepts {compiler_flags} and linker accepts {linker_flags}")
            return False
        if not binary.exists() or not binary.is_file() or not access(binary, R_OK):
            logger.info(f"Compilation did not produce an executable output file, check that {target.compiler} works")
            return False

    return True


@cache
def simple_instr_compiles(which: str) -> bool:
    from subprocess import CalledProcessError
    if not check_for_mccode_antlr_compiler(which):
        return False
    try:
        from mccode_antlr.loader import parse_mcstas_instr
        instr = parse_mcstas_instr("define instrument check() trace component a = Arm() at (0,0,0) absolute end")
        return compiles(which, instr)
    except RuntimeError:
        return False
    except FileNotFoundError:
        return False
    except CalledProcessError:
        return False


def compiled(method, compiler: str | None = None):
    from unittest import TestCase
    if compiler is None:
        # Basic compiled instruments only need the 'cc' compiler specified in the config file
        compiler = 'cc'

    def wrapper(*args, **kwargs):
        if simple_instr_compiles(compiler):
            return method(*args, **kwargs)
        elif isinstance(args[0], TestCase):
            args[0].skipTest(f'Skipping due to lack of working ${compiler}')
        else:
            raise RuntimeError(f'A working compiler is required to use function {method.__name__}')

    return wrapper


def gpu_only(method):
    from loguru import logger
    # GPU compiled instruments need the specific OpenACC compiler
    # **PLUS** they need to _actually_ have the openACC header (macOS and Windows don't use different compilers)
    return compiled(method, 'acc')


def mpi_only(method):
    # MPI compiled instruments need the specified compiler
    return compiled(method, 'mpi/cc')



from unittest import TestCase
from .utils import make_assembler


class TestInstrEmptyTrace(TestCase):
    def test_parse_empty_trace(self):
        from mccode_antlr.instr import Instr
        from mccode_antlr.common import InstrumentParameter, Expr, Value, DataType
        from mccode_antlr.loader import parse_mcstas_instr
        instr_source = """
        DEFINE INSTRUMENT test_parse(par0=3.14159, double par1 = 49, int par2 =     1010110
            , string par3="this is a long string with spaces", 

            string par4, int par5, double par6, par7)
        TRACE
        END
        """
        instr = parse_mcstas_instr(instr_source)
        self.assertTrue(isinstance(instr, Instr))
        self.assertEqual(instr.name, 'test_parse')
        self.assertEqual(len(instr.parameters), 8)
        self.assertEqual(len(instr.components), 0)
        for i, (d_type, val) in enumerate([('double', 3.14159), ('double', 49), ('int', 1010110),
                                           ('string', '"this is a long string with spaces"'),
                                           ('string', None), ('int', None), ('double', None), ('double', None)]):
            p = instr.parameters[i]
            self.assertTrue(isinstance(p, InstrumentParameter))
            self.assertEqual(p.name, f'par{i}')
            self.assertTrue(isinstance(p.value, Expr))
            self.assertEqual(p.value, Value(val, data_type=DataType.from_name(d_type)))

    def test_assemble_empty_trace(self):
        from mccode_antlr.instr import Instr
        from mccode_antlr.common import InstrumentParameter, Expr, Value, DataType

        assembler = make_assembler('test_assemble')

        # There are multiple ways to specify instrument parameters:
        # 1. specify a single parameter object to add:
        assembler.parameter(InstrumentParameter('par0', '', Expr.float(3.14159)))
        # 2. specify a single parameter as a string to be parsed:
        assembler.parameter('double par1 = 49')
        # 3. Using the parameters method, specify any number of parameter objects or strings to be parsed (as above)
        #    and/or specify keyword value pairs, where each value can be
        #      a. a two-element list or tuple of 'value', 'unit'
        #      b. a dict with entries 'unit' and 'value'
        #      c. a plain value
        #    in each of these cases, Expr.best converts the provided value if it is not already an Expr object
        assembler.parameters(InstrumentParameter('par2', '', Expr.int(1010110)),
                             'string par3 = "this is a long string with spaces"',
                             InstrumentParameter('par4', '', Expr.str('"the-fourth_parameter"')),
                             'int par5=-9', par6=[Expr.float(None), ''], par7=191, par8=dict(unit='', value='"whoa!"'))

        instr = assembler.instrument
        self.assertTrue(isinstance(instr, Instr))
        self.assertEqual(instr.name, 'test_assemble')
        self.assertEqual(len(instr.parameters), 9)
        self.assertEqual(len(instr.components), 0)
        d_type_vals = [('double', 3.14159), ('double', 49), ('int', 1010110),
                       ('string', '"this is a long string with spaces"'), ('string', '"the-fourth_parameter"'),
                       ('int', -9), ('float', None), ('int', 191), ('string', '"whoa!"')]
        for i, (par, (d_type, val)) in enumerate(zip(instr.parameters, d_type_vals)):
            self.assertTrue(isinstance(par, InstrumentParameter))
            self.assertEqual(par.name, f'par{i}')
            self.assertTrue(isinstance(par.value, Expr))
            self.assertEqual(par.value, Value(val, data_type=DataType.from_name(d_type)))



from unittest import TestCase
from .utils import parse_instr_string


class TestInstrParameters(TestCase):

    def test_tas1_c1_parameter(self):
        from mccode_antlr.loader.loader import parse_mccode_instr_parameters
        contents = """ DEFINE INSTRUMENT tas(PHM=-37.077,TTM=-74,C1=30) TRACE END"""
        instr_parameters = parse_mccode_instr_parameters(contents)
        self.assertEqual(len(instr_parameters), 3)
        self.assertEqual(instr_parameters[0].name, 'PHM')
        self.assertEqual(instr_parameters[1].name, 'TTM')
        self.assertEqual(instr_parameters[2].name, 'C1')
        self.assertEqual(instr_parameters[0].value, -37.077)
        self.assertEqual(instr_parameters[1].value, -74)
        self.assertEqual(instr_parameters[2].value, 30)

    def test_used_parameter_check(self):
        from mccode_antlr.assembler import Assembler
        from mccode_antlr.reader import MCSTAS_REGISTRY
        assembler = Assembler('test_used_parameter_check', registries=[MCSTAS_REGISTRY])
        assembler.parameter('double par0/"pi" = 3.14159')
        assembler.parameter('int par1 = 49')
        assembler.parameter('int par2 = 1010110')
        assembler.parameter('string par3 = "this is a long string with spaces"')
        assembler.parameter('string par4 = "the-fourth_parameter"')
        assembler.parameter('par5 = -9.11')
        assembler.parameter('a4/"degree" = 90')
        assembler.component('origin', 'Progress_bar', at=(0, 0, 0))
        assembler.component('left', 'PSD_monitor', at=(0, 0, 1), parameters=dict(
            xwidth='par0', yheight='2*fmod(par5, 0.1)', nx='par1', ny='par2 >> 4', restore_neutron=1))
        # TODO (maybe) the right shift operator is not implemented, instead the expression is parsed as
        #      BinOp('>', [BinOp('>', ['par2', '']), '4']) {or similar} which passes this test but may not
        #      be sufficient for all cases
        assembler.component('up', 'PSD_monitor', at=(0, 0, 2), parameters=dict(
            xwidth='cos(DEG2RAD * a4)', yheight='sin(DEG2RAD * a4)', nx='-a4 * DEG2RAD', restore_neutron=1))

        inst = assembler.instrument
        self.assertEqual(7, len(inst.parameters))
        self.assertEqual(3, len(inst.components))
        self.assertTrue(inst.parameter_used('par0'))
        self.assertTrue(inst.parameter_used('par1'))
        self.assertTrue(inst.parameter_used('par2'))
        self.assertFalse(inst.parameter_used('par3'))
        self.assertFalse(inst.parameter_used('par4'))
        self.assertTrue(inst.parameter_used('par5'))
        self.assertTrue(inst.parameter_used('a4'))
        self.assertEqual(2, inst.check_instrument_parameters(), )



def cogen_display(source, declared_parameters):
    lines = ["/* *****************************************************************************",
             f"* instrument '{source.name}' and components DISPLAY",
             "***************************************************************************** */",
             ]
    macros = dict(magnify='mcdis_magnify', line='mcdis_line', dashed_line='mcdis_dashed_line',
                  multiline='mcdis_multiline', rectangle='mcdis_rectangle', box='mcdis_box',
                  circle='mcdis_circle', cylinder='mcdis_cylinder', sphere='mcdis_sphere',
                  cone='mcdis_cone', polygon='mcdis_polygon', polyhedron='mcdis_polyhedron',
                  )

    for x, v in macros.items():
        lines.append(f'  #define {x} {v}')

    for comp in source.component_types():
        lines.extend(cogen_comp_display_class(comp, declared_parameters[comp.name]))

    for x in macros:
        lines.append(f'  #undef {x}')

    # write the instrument main code, which calls component ones
    lines.append(f'int display(void) {{ /* called by mccode_main for {source.name}:DISPLAY */')
    # lines.extend([f"#pragma acc update host(_{comp.name}_var)" for comp in source.components])
    lines.extend([
        '  printf("MCDISPLAY: start\\n");',
        '',
    ])

    # Of course, there is no instrument level DISPLAY block directive
    # # insert user code from instrument definition
    # if len(source.display):
    #     f, n = source.display[0].fn
    #     lines.extend([
    #         f'  /* Instrument {source.name} DISPLAY */',
    #         f'  SIG_MESSAGE("[{source.name} DISPLAY [{f}:{n}]");'
    #     ])
    #     for par in source.parameters:
    #         # ensure there's no conflict of names
    #         lines.append(f'  #define {par.name} (instrument->_parameters.{par.name}')
    #     for block in source.display:
    #         lines.append(block.to_c())
    #     for par in source.parameters:
    #         lines.append(f'  #undef {par.name}')

    lines.append('  /* call iteratively all components DISPLAY */')
    for comp in source.components:
        if len(comp.type.display):
            lines.append(f'  class_{comp.type.name}_display(&_{comp.name}_var);')
        else:
            # McCode-3 would have called a function just to print this line
            lines.append(f'  printf("MCDISPLAY: component %s\\n", _{comp.name}_var._name);')

    lines.extend([
        '  printf("MCDISPLAY: end\\n");'
        '',
        '  return(0);',
        '} /* display */',
        ''
    ])
    return '\n'.join(lines)


def cogen_comp_display_class(comp, declared_parameters):
    from .c_defines import cogen_parameter_define, cogen_parameter_undef
    if not len(comp.display):
        return []

    lines = [
        # f'_class_{comp.name} *class_{comp.name}_display(_class_{comp.name} *_comp) {{',
        # Changed following https://github.com/McStasMcXtrace/McCode/commit/73a32858af1a5193bbb692da49379816cb0f9e38
        f'void class_{comp.name}_display(_class_{comp.name} *_comp) {{',
        cogen_parameter_define(comp, declared_parameters)
    ]
    f, n = comp.display[0].fn if len(comp.final) else (comp.name, 0)
    lines.append(f'  SIG_MESSAGE("[_{comp.name}_display] component NULL={comp.name}() [{f}:{n}]");')
    lines.append('  printf("MCDISPLAY: component %s\\n", _comp->_name);')

    for block in comp.display:
        lines.append(block.to_c())

    lines.extend([
        cogen_parameter_undef(comp, declared_parameters),
        # '  return(_comp);',
        '  return;',
        f'}} /* class_{comp.name}_display */',
        ''
    ])
    return lines



from dataclasses import dataclass, field
from typing import Union
from pathlib import Path


def read_keywords(lines: list[str], keywords: list[str], delim: str = None):
    if delim is None:
        delim = ':'
    if len(lines) < len(keywords):
        raise RuntimeError(f"Expected {len(keywords)} lines, got {len(lines)}")
    for line, keyword in zip(lines, keywords):
        key, _, value = line.partition(delim)
        if key.strip() != keyword:
            raise RuntimeError(f"Expected keyword {keyword}, got {key}")
        yield value.strip()


@dataclass
class SimFileHeader:
    description: str
    Date: str
    Program: str

    @staticmethod
    def parse(contents: str):
        contents = contents.strip()
        lines = contents.split('\n')
        if len(lines) != 3:
            raise RuntimeError(f"Expected 3 lines in header, got {len(lines)}")
        description = lines[0].strip()
        date, program = read_keywords(lines[1:], ['Date', 'Program'])
        return SimFileHeader(description=description, Date=date, Program=program)

    def to_file(self, file):
        print(self.description, file=file)
        print(f"Date: {self.Date}", file=file)
        print(f"Program: {self.Program}", file=file)

    def __contains__(self, other):
        if not isinstance(other, SimFileHeader):
            raise RuntimeError(f"Cannot compare header with {other}")
        return other.description in self.description and other.Date in self.Date and other.Program in self.Program

    def combine(self, other):
        if other in self:
            return self
        if self in other:
            return other
        # combine from the right, so self.description might already have other.description in it
        desc = self.description if other.description in self.description else f'{self.description} + {other.description}'
        date = self.Date if other.Date in self.Date else f"{self.Date} + {other.Date}"
        program = self.Program if other.Program in self.Program else f"{self.Program} + {other.Program}"
        return SimFileHeader(description=desc, Date=date, Program=program)


@dataclass
class SimFileInstrument:
    name: str
    File: str
    Source: str
    Parameters: str
    Trace_enabled: bool
    Default_main: bool
    Embedded_runtime: bool

    @staticmethod
    def parse(contents: str):
        contents = contents.strip()
        lines = contents.split('\n')
        if len(lines) != 7:
            raise RuntimeError(f"Expected 7 lines in instrument, got {len(lines)}")
        name = lines[0].strip()
        keywords = ['File', 'Source', 'Parameters', 'Trace_enabled', 'Default_main', 'Embedded_runtime']
        file, source, parameters, trace, default, embed = read_keywords(lines[1:], keywords)
        trace = trace == 'yes'
        default = default == 'yes'
        embed = embed == 'yes'
        return SimFileInstrument(name=name, File=file, Source=source, Parameters=parameters, Trace_enabled=trace,
                                 Default_main=default, Embedded_runtime=embed)

    def to_file(self, file):
        print(f'begin instrument: {self.name}', file=file)
        print(f"  File: {self.File}", file=file)
        print(f"  Source: {self.Source}", file=file)
        print(f"  Parameters: {self.Parameters}", file=file)
        print(f"  Trace_enabled: {'yes' if self.Trace_enabled else 'no'}", file=file)
        print(f"  Default_main: {'yes' if self.Default_main else 'no'}", file=file)
        print(f"  Embedded_runtime: {'yes' if self.Embedded_runtime else 'no'}", file=file)
        print(f'end instrument', file=file)

    def __contains__(self, other):
        if not isinstance(other, SimFileInstrument):
            raise RuntimeError(f"Cannot compare instrument with {other}")
        return other.name in self.name and other.File in self.File and other.Source in self.Source and \
            other.Parameters in self.Parameters and other.Trace_enabled == self.Trace_enabled and \
            other.Default_main == self.Default_main and other.Embedded_runtime == self.Embedded_runtime

    def combine(self, other):
        if other in self:
            return self
        if self in other:
            return other
        # combine from the right, so self.name might already have other.name in it
        name = self.name if other.name in self.name else f'{self.name} + {other.name}'
        file = self.File if other.File in self.File else f"{self.File} + {other.File}"
        source = self.Source if other.Source in self.Source else f"{self.Source} + {other.Source}"
        # parameters are a bit different in that we want their ordered union:
        pars = dict.fromkeys([p.strip() for p in self.Parameters.split(' ')])
        pars.update(dict.fromkeys([p.strip() for p in other.Parameters.split(' ')]))
        parameters = ' '.join(list(pars.keys()))
        trace = self.Trace_enabled and other.Trace_enabled
        default = self.Default_main and other.Default_main
        embed = self.Embedded_runtime and other.Embedded_runtime
        return SimFileInstrument(name=name, File=file, Source=source, Parameters=parameters, Trace_enabled=trace,
                                 Default_main=default, Embedded_runtime=embed)


@dataclass
class SimFileSimulation:
    name: str
    Format: str
    URL: str
    Creator: str
    Instrument: str
    Ncount: int
    Trace: bool
    Gravitation: bool
    Seed: int
    Directory: str
    Params: list[str] = field(default_factory=list)

    @staticmethod
    def parse(contents: str):
        contents = contents.strip()
        lines = contents.split('\n')
        if len(lines) < 10:
            raise RuntimeError(f"Expected at least 10 lines in simulation, got {len(lines)}")
        name = lines[0].strip()
        keywords = ['Format', 'URL', 'Creator', 'Instrument', 'Ncount', 'Trace', 'Gravitation', 'Seed', 'Directory']
        format, url, creator, instrument, ncount, trace, gravitation, seed, directory = read_keywords(lines[1:10],
                                                                                                      keywords)
        params = []
        for line in lines[10:]:
            _, _, param = line.partition(':')
            params.append(param.strip())
        ncount = int(ncount)
        trace = trace == 'yes'
        gravitation = gravitation == 'yes'
        seed = int(seed)
        return SimFileSimulation(name=name, Format=format, URL=url, Creator=creator, Instrument=instrument,
                                 Ncount=ncount, Trace=trace, Gravitation=gravitation, Seed=seed, Directory=directory,
                                 Params=params)

    def to_file(self, file):
        print(f'begin simulation: {self.name}', file=file)
        print(f"  Format: {self.Format}", file=file)
        print(f"  URL: {self.URL}", file=file)
        print(f"  Creator: {self.Creator}", file=file)
        print(f"  Instrument: {self.Instrument}", file=file)
        print(f"  Ncount: {self.Ncount}", file=file)
        print(f"  Trace: {'yes' if self.Trace else 'no'}", file=file)
        print(f"  Gravitation: {'yes' if self.Gravitation else 'no'}", file=file)
        print(f"  Seed: {self.Seed}", file=file)
        print(f"  Directory: {self.Directory}", file=file)
        for param in self.Params:
            print(f"  Param: {param}", file=file)
        print(f'end simulation', file=file)

    def __contains__(self, other):
        if not isinstance(other, SimFileSimulation):
            raise RuntimeError(f"Cannot compare simulation with {other}")
        return other.name in self.name and other.Format in self.Format and other.URL in self.URL and \
            other.Creator in self.Creator and other.Instrument in self.Instrument and \
            other.Trace == self.Trace and other.Gravitation == self.Gravitation and \
            other.Directory in self.Directory and all(p in self.Params for p in other.Params)

    def __eq__(self, other):
        if not isinstance(other, SimFileSimulation):
            raise RuntimeError(f"Cannot compare simulation with {other}")
        return other.name == self.name and other.Format == self.Format and other.URL == self.URL and \
            other.Creator == self.Creator and other.Instrument == self.Instrument and \
            other.Trace == self.Trace and other.Gravitation == self.Gravitation and \
            other.Directory == self.Directory and all(p in self.Params for p in other.Params)

    def combine(self, other):
        from copy import deepcopy
        if other == self:
            # combining a repeat which was done to improve statistics
            out = deepcopy(self)
            out.Ncount += other.Ncount
            if other.Seed == self.Seed:
                from loguru import logger
                logger.warning('combining two simulations with the same seed is not statistically valid')
            return out
        if other in self:
            return self
        elif self in other:
            return other
        # combine from the right, so self.name might already have other.name in it
        name = self.name if other.name in self.name else f'{self.name} + {other.name}'
        format = self.Format if other.Format in self.Format else f"{self.Format} + {other.Format}"
        url = self.URL if other.URL in self.URL else f"{self.URL} + {other.URL}"
        creator = self.Creator if other.Creator in self.Creator else f"{self.Creator} + {other.Creator}"
        instrument = self.Instrument if other.Instrument in self.Instrument else f"{self.Instrument} + {other.Instrument}"
        # TODO this is *not* a repeat. What do we do with the number of particles?
        ncount = min(self.Ncount, other.Ncount)
        trace = self.Trace and other.Trace
        gravitation = self.Gravitation and other.Gravitation
        seed = self.Seed if self.Seed == other.Seed else -1
        directory = self.Directory if other.Directory in self.Directory else f"{self.Directory} + {other.Directory}"
        params = dict.fromkeys(self.Params)
        params.update(dict.fromkeys(other.Params))
        params = list(params.keys())
        return SimFileSimulation(name=name, Format=format, URL=url, Creator=creator, Instrument=instrument,
                                 Ncount=ncount, Trace=trace, Gravitation=gravitation, Seed=seed, Directory=directory,
                                 Params=params)


@dataclass
class SimFileData:
    keywords: list[str]
    Date: str
    type: str
    Source: str
    component: str
    position: str
    title: str
    Ncount: int
    statistics: str
    signal: str
    values: str
    filename: str = None  # Not defined for 0-D data entries
    xvar: str = None  # Not defined for 0-D data entries
    yvar: str = None  # Not defined for 0-D data entries
    xlabel: str = None  # Not defined for 0-D data entries
    ylabel: str = None  # Not defined for 0-D data entries
    variables: str = None  # Not defined for 0-D data entries
    xlimits: str = None  # only defined for 1-D data entries
    zvar: str = None  # only defined for 2-D data entries
    zlabel: str = None  # only defined for 2-D data entries
    xylimits: str = None  # only defined for 2-D data entries

    @staticmethod
    def parse(contents: str):
        contents = contents.strip()
        lines = contents.split('\n')
        first_keywords = ['Date', 'type', 'Source', 'component', 'position', 'title', 'Ncount']
        second_keywords = ['statistics', 'signal', 'values']
        file_keyword = []
        if len(lines) == 12:
            # 0-D -- old style
            last_keywords = ['xylimits', 'variables']
        elif len(lines) == 13:
            # 0-D -- new style (as of when?)
            file_keyword = ['filename']
            last_keywords = ['xylimits', 'variables']
        elif len(lines) == 17:
            # 1-D
            file_keyword = ['filename']
            last_keywords = ['xvar', 'yvar', 'xlabel', 'ylabel', 'xlimits', 'variables']
        elif len(lines) == 19:
            # 2-D
            file_keyword = ['filename']
            last_keywords = ['xvar', 'yvar', 'xlabel', 'ylabel', 'zvar', 'zlabel', 'xylimits', 'variables']
        else:
            raise RuntimeError(f"Expected 12, 13, 17 or 19 lines in data, got {len(lines)}")
        keywords = first_keywords + file_keyword + second_keywords + last_keywords
        values = list(read_keywords(lines, keywords))
        collected = {k: v for k, v in zip(keywords, values)}
        collected['Ncount'] = int(collected['Ncount'])
        return SimFileData(keywords=keywords, **collected)

    def to_file(self, file):
        print(f'begin data', file=file)
        for key in self.keywords:
            if getattr(self, key) is not None:
                print(f'  {key}: {getattr(self, key)}', file=file)
        print(f'end data', file=file)

    def __eq__(self, other):
        if not isinstance(other, SimFileData):
            raise RuntimeError(f"Cannot compare data with {other}")
        return all(getattr(self, k) == getattr(other, k) for k in self.keywords)

    def is_repeat(self, other):
        # Everything must match except Date, Ncount, statistics, signal, values.
        # The Date is the date the simulation was run, so it will be different.
        # The Ncount is the number of particles used in the simulation, so it can be different.
        # The statistics, signal, and values are the results of the simulation, so they will be different.
        if not isinstance(other, SimFileData):
            raise RuntimeError(f"Cannot compare data with {other}")
        skip = ['Date', 'Ncount', 'statistics', 'signal', 'values']
        return all(getattr(self, k) == getattr(other, k) for k in self.keywords if k not in skip)

    def combine(self, other):
        from math import sqrt
        # this should only happen if the simulations are repeats of each other
        if not self.is_repeat(other):
            raise RuntimeError(f"Cannot combine non-repeated {self} with {other}")
        from copy import deepcopy
        out = deepcopy(self)
        out.Ncount += other.Ncount

        def split_statistics(s: str):
            return {k: float(v) for k, v in [z.strip(';').split('=', 1) for z in s.split()]}

        def split_signal(s: str):
            return [float(x.split('=')[1].split(';')[0]) for x in s.split(' ')]

        def split_values(v: str):
            return [int(x) if x.is_integer() else x for x in [float(x) for x in v.split(' ')]]

        # The values line is always "{float} {float} {int}" which _I think_ is <intensity>, <error>, and event count
        s_intensity, s_error, s_counts = split_values(self.values)
        o_intensity, o_error, o_counts = split_values(other.values)
        out.values = f"{s_intensity + o_intensity} {sqrt(s_error * s_error + o_error + o_error)} {s_counts + o_counts}"

        # The signal line is always? "None" or "Min={float}; Max={float}; Mean={float};"
        # we can combine min and max, but not mean (without knowing the number of points)
        if 'None' in self.signal or 'None' in other.signal:
            out.signal = 'None'
        else:
            s_min, s_max, s_mean = split_signal(self.signal)
            o_min, o_max, o_mean = split_signal(other.signal)
            mean_estimate = (s_mean * s_counts + o_mean * o_counts)/(s_counts + o_counts)
            out.signal = f"Min={min(s_min, o_min)}; Max={max(s_max, o_max)}; Mean={mean_estimate};"

        # Statistics: one of "None", "X0={float}; dX={float};" or "X0={float}; dX={float}; Y0={float}; dY={float};"
        # we can combine the values under the assumption that the distribution is Gaussian, but only if we know
        # the overall intensity as well.
        if 'None' in self.statistics or 'None' in other.statistics:
            out.statistics = 'None'
        else:
            s_stats = split_statistics(self.statistics)
            o_stats = split_statistics(other.statistics)
            stats = {}
            for d in ('X', 'Y'):
                d0, dd = f'{d}0', f'd{d}'
                if d0 in s_stats and d0 in o_stats:
                    stats[d0] = (s_stats[d0] * s_counts + o_stats[d0] * o_counts)/(s_counts + o_counts)
                if dd in s_stats and dd in o_stats:
                    sd = s_stats[dd] * s_counts
                    od = o_stats[dd] * o_counts
                    stats[dd] = sqrt(sd * sd + od * od) / (s_counts + o_counts)
            out.statistics = ' '.join([f"{k}={stats[k]};" for k in ['X0', 'dX', 'Y0', 'dY'] if k in stats])
        return out


@dataclass
class SimFile:
    header: SimFileHeader = field(default_factory=SimFileHeader)
    instrument: SimFileInstrument = field(default_factory=SimFileInstrument)
    simulation: SimFileSimulation = field(default_factory=SimFileSimulation)
    data: list[SimFileData] = field(default_factory=list)

    @staticmethod
    def load(source: Union[str, Path]):
        if not isinstance(source, Path):
            source = Path(source)
        if not source.is_file():
            raise FileNotFoundError(f"Could not find file {source}")
        with source.open('r') as f:
            contents = f.read()
        return SimFile.parse(contents)

    def save(self, filename: Union[str, Path]):
        if not isinstance(filename, Path):
            filename = Path(filename)
        with filename.open('w') as f:
            self.to_file(f)

    @staticmethod
    def parse(contents: str):
        # split into sub-string parts:
        header, _, rest = contents.partition('begin instrument: ')
        instrument, _, rest = rest.partition('begin simulation: ')
        instrument, _, _ = instrument.partition('end instrument')
        simulation, _, rest = rest.partition('end simulation')
        data = []
        while rest:
            datum, _, rest = rest.partition('end data')
            _, _, datum = datum.partition('begin data')
            if len(datum.strip()):
                data.append(datum.strip())

        # convert sub-strings into objects:
        header = SimFileHeader.parse(header)
        instrument = SimFileInstrument.parse(instrument)
        simulation = SimFileSimulation.parse(simulation)
        data = [SimFileData.parse(datum.strip()) for datum in data]
        return SimFile(header=header, instrument=instrument, simulation=simulation, data=data)

    def to_file(self, file):
        self.header.to_file(file)
        print(file=file)
        self.instrument.to_file(file)
        print(file=file)
        self.simulation.to_file(file)
        print(file=file)
        for datum in self.data:
            datum.to_file(file)
            print(file=file)


def read_mccode_sim(filename: Union[str, Path]):
    return SimFile.load(filename)


def combine_mccode_sims(sims: list[SimFile]):
    """Combine multiple McCode simulation files together into a single file.

    Each file could be a repeat of the same simulation (to improve statistics)
    or an extension of an earlier simulation to add flightpath, components, detectors, etc."""
    if len(sims) == 0:
        raise RuntimeError("No simulations to combine")

    for sim in sims:
        if not isinstance(sim.header, SimFileHeader):
            raise RuntimeError(f"Header not a SimFileHeader")
        if not isinstance(sim.instrument, SimFileInstrument):
            raise RuntimeError(f"Instrument not a SimFileInstrument")
        if not isinstance(sim.simulation, SimFileSimulation):
            raise RuntimeError(f"Simulation is not a SimFileSimulation")
        for datum in sim.data:
            if not isinstance(datum, SimFileData):
                raise RuntimeError(f"Datum not a SimFileData")

    header = sims[0].header
    instrument = sims[0].instrument
    simulation = sims[0].simulation
    data = {(d.component, d.filename): d for d in sims[0].data}
    for sim in sims[1:]:
        header = header.combine(sim.header)
        instrument = instrument.combine(sim.instrument)
        simulation = simulation.combine(sim.simulation)
        for datum in sim.data:
            key = (datum.component, datum.filename)
            if key not in data:
                data[key] = datum
            else:
                data[key] = data[key].combine(datum)
    return SimFile(header=header, instrument=instrument, simulation=simulation, data=list(data.values()))


def write_combined_mccode_sims(files: list[Union[Path, str]], output: Union[Path, str]):
    sims = [SimFile.load(f) for f in files]
    combined = combine_mccode_sims(sims)
    combined.save(output)



from io import StringIO
from ..instr import Instr, Instance
from ..comp import Comp
from loguru import logger

MCSTAS_GENERATOR = dict(project=1, name="mcstas", fancy="McStas", url='http://www.mcstas.org')
MCXTRACE_GENERATOR = dict(project=2, name='mcxtrace', fancy='McXtrace', url='http://www.mcxtrace.org')

# These _should_ be set in a call to, e.g., `mcstas4`,
# They were previously set in the yacc generated main function
CONFIG = dict(default_main=True, enable_trace=True, portable=True, include_runtime=True,
              embed_instrument_file=False,)


# Follow the logic of codegen.c(.in) from McCode-3, but make use of visitor semantics for possible alternate runtimes
class TargetVisitor:
    def __init__(self, instr: Instr, generate: dict = None, config: dict = None, verbose=False):
        self.runtime = MCSTAS_GENERATOR if generate is None else generate
        self.config = CONFIG if config is None else config
        self.source = instr
        self.output = None
        self.verbose = verbose
        self.warnings = 0
        self.instrument_uservars = ()
        self.component_uservars = dict()
        self.includes = []
        self.typedefs = None
        self.component_declared_parameters = dict()
        self.ok_to_skip = None
        #
        self.source.verify_instance_parameters()
        self.__post_init__()

    def __init_subclass__(cls, **kwargs):
        target_language = kwargs.get('target_language', 'none')
        if 'target_language' in kwargs:
            del kwargs['target_language']
        super().__init_subclass__(**kwargs)
        cls.target_language = target_language

    def __post_init__(self):
        pass

    @property
    def is_mcstas(self):
        index = self.runtime.get('project')
        if index != 1 and index != 2:
            raise RuntimeError(f'Unknown runtime for project index {index}')
        return index == 1

    @property
    def registries(self):
        # TODO always ensure this is sorted by priority?
        return self.source.registries

    def known(self, name: str, which: str = None):
        if self.registries is None:
            return False
        registries = self.registries if which is None else [x for x in self.registries if x.name in which]
        return any([reg.known(name) for reg in registries])

    def locate(self, name: str, which: str = None):
        registries = self.registries if which is None else [x for x in self.registries if x.name in which]
        for reg in registries:
            if reg.known(name):
                return reg.path(name)
        names = [reg.name for reg in registries]
        msg = "registry " + names[0] if len(names) == 1 else 'registries: ' + ','.join(names)
        raise RuntimeError(f'{name} not found in {msg}')

    def library_path(self, filename=None):
        return self.locate(filename)

    def configure_file(self, filename: str, flavor: str):
        """McStasMcXtrace/McCode makes use of CMake configure_file to define runtime-header macros.
        These file(s) have the extension .h.in and are changed to .h after replacing all "@MCCODE_*@ keys
        """
        import re
        from importlib.resources import as_file
        from mccode_antlr.config import config, registry_defaults
        if not filename.endswith(".h.in"):
            raise RuntimeError(f"Expected to configure only header files, not {filename}")
        with as_file(self.library_path(filename)) as file_at:
            not_configured_name = str(file_at)
            with open(file_at, 'r') as file:
                contents = file.read()

        # It's not great to do this here. TODO Find a better place for this
        reg = [reg for reg in self.registries if reg.unique(filename)]
        if len(reg) != 1:
            raise RuntimeError(f"Expected exactly one registry for file {filename}")
        # updates mccode_antlr.config.config
        registry_defaults(reg[0], [reg[0].name])

        def replacement(match) -> str:
            name = match.group(1).lower()
            key = config[flavor][name]
            if key.exists():
                return str(key.get())  # key might not represent a string, but we need to return one
            if 'date' in name:
                return '1970-01-01'
            if name == 'version_macro' or 'number' in name:
                return '0'
            return "{UNKNOWN}"

        # find all occurances of @MCCODE_([A-Z]*)@, replace them with their equivalent config[flavor][$1].get()
        contents = re.sub(r'@MCCODE_([A-Z_]*)@', replacement, contents)

        # one could consider writing the configured contents to a (temporary) file for use if the translator
        # is not set to 'include_runtime'
        self.out(f'/* embedding configured version of file "{not_configured_name}" */')
        self.out(contents)
        self.out(f'/* end of configured version of file "{not_configured_name}" */')

    def embed_file(self, filename):
        """Reads the library file, even if embedded in a module archive, writes to the output IO Stream"""
        from importlib.resources import as_file
        with as_file(self.library_path(filename)) as file_at:
            with open(file_at, 'r') as file:
                self.out(f'/* embedding file "{file_at}" */')
                self.output.write(file.read())
                self.out(f'/* end of file "{file_at}" */')

    def include_path(self, filename=None):
        if filename is None:
            # It would be better to read this from the cache itself
            from pooch import os_cache
            return os_cache('mccode_antlr-libc')
        return self.library_path(filename)

    def out(self, value):
        if self.output is None:
            raise RuntimeError('Printing only enabled once translation has begun!')
        print(value, file=self.output)

    def info(self, msg: str):
        if self.verbose:
            logger.info(f'{self.source.name}: {msg}')

    def translate(self, reprocess=True):
        if self.output is not None:
            if not reprocess:
                return self.output
            self.output.close()
        self.output = StringIO()
        self.info('visit header')
        self.visit_header()
        self.info('visit declare')
        self.visit_declare()
        self.info('set jump targets')
        self.set_jump_absolute_targets()
        self.info('determine transforms to skip')
        self.detect_skipable_transforms()
        self.info('visit initialize')
        self.visit_initialize()
        self.info('enter trace')
        self.enter_trace()
        self.info('enter uservars')
        self.enter_uservars()
        self.info('visit raytrace')
        self.visit_raytrace()
        self.info('visit raytrace funnel')
        self.visit_raytrace_funnel()
        self.info('leave uservars')
        self.leave_uservars()
        self.info('leave trace')
        self.leave_trace()
        self.info('visit save')
        self.visit_save()
        self.info('visit finally')
        self.visit_finally()
        self.info('visit display')
        self.visit_display()
        self.info('visit macros')
        self.visit_macros()
        if self.verbose and self.warnings:
            print(f"Build of instrument {self.source.name} had {self.warnings} warnings")
        return self.output

    def save(self, filename=None, close=True, reprocess=True):
        self.translate(reprocess=reprocess)
        if filename:
            with open(filename, 'w') as file:
                file.write(self.output.getvalue())
        if close:
            self.output.close()
            self.output = None

    def contents(self):
        if self.output is None:
            output = self.translate(reprocess=False)
            string = output.getvalue()
            output.close()
            self.output = None
        else:
            string = self.output.getvalue()
        return string

    def detect_skipable_transforms(self):
        def can_skip(comp):
            # Any component with a TRACE or EXTEND can not be skipped
            if any(not x.is_empty for x in comp.extend) or any(not x.is_empty for x in comp.type.trace):
                return False
            # Any component jumping elsewhere can not be skipped
            if len(comp.jump):
                return False
            return True

        can_skip_transform = [can_skip(comp) for comp in self.source.components]

        # Any component that is jumped *to* can not be skipped (set_jump_absolute_targets must be called before this)
        for jump in [j for c in self.source.components for j in c.jump]:
            can_skip_transform[jump.absolute_target] = False

        self.ok_to_skip = can_skip_transform

    def set_jump_absolute_targets(self):
        for index, jump in [(i, j) for i, c in enumerate(self.source.components) for j in c.jump]:
            # jump is a dataclass with 'target', 'index', 'iterate', 'condition', and 'actual_target_index'
            if jump.absolute_target > -1:
                # Somewhere/something already set the actual target index -- we don't have a choice but to trust it
                continue
            if jump.relative_target == 0 and jump.target.lower() != 'myself':
                # the target is another named component:
                jump.absolute_target = [i for i, c in enumerate(self.source.components) if jump.target == c.name][0]
                jump.relative_target = jump.absolute_target - index
            else:
                jump.absolute_target = index + jump.relative_target

    def enter_trace(self):
        pass
            
    def leave_trace(self):
        pass

    def visit_header(self):
        pass

    def visit_declare(self):
        pass

    def visit_initialize(self):
        # likely target dependent
        pass

    def enter_uservars(self):
        pass
    
    def leave_uservars(self):
        pass
    
    def visit_raytrace(self):
        pass
    
    def visit_raytrace_funnel(self):
        pass

    def visit_save(self):
        pass

    def visit_finally(self):
        pass

    def visit_display(self):
        pass

    def visit_macros(self):
        pass

        
        




from unittest import TestCase


def _make_seitz_list(tx, ty, tz, degrees=False):
    from mccode_antlr.common import Expr
    from mccode_antlr.instr.orientation import Seitz, cos_value, sin_value
    o, z = Expr.float(1), Expr.float(0)
    cx, sx = cos_value(tx, degrees), sin_value(tx, degrees)
    cy, sy = cos_value(ty, degrees), sin_value(ty, degrees)
    cz, sz = cos_value(tz, degrees), sin_value(tz, degrees)
    rx = Seitz(o, z, z, z, z, cx, sx, z, z, -sx, cx, z)
    ry = Seitz(cy, z, -sy, z, z, o, z, z, sy, z, cy, z)
    rz = Seitz(cz, sz, z, z, -sz, cz, z, z, z, z, o, z)
    return rx, ry, rz


def _random_angles_radian():
    from numpy import random, pi
    from mccode_antlr.common import Expr
    return [Expr.float(random.rand() * 2 * pi - pi) for _ in range(3)]


def _random_angles_degrees():
    from numpy import random
    from mccode_antlr.common import Expr
    return [Expr.float(random.rand() * 360 - 180) for _ in range(3)]


def _random_vector(minimum: float = 0, maximum: float = 1):
    from numpy import random
    from mccode_antlr.common import Expr
    from mccode_antlr.instr.orientation import Vector

    def _random():
        return Expr.float(random.rand() * (maximum - minimum) + minimum)

    return Vector(_random(), _random(), _random())


class TestOrientation(TestCase):
    def test_matrix_vector_multiply(self):
        from mccode_antlr.instr.orientation import Vector, Matrix
        from mccode_antlr.common import Expr
        from numpy import random
        v = Vector(*[Expr.float(random.rand()) for _ in range(3)])
        m = Matrix(*[Expr.float(random.rand()) for _ in range(9)])
        mv = m * v
        self.assertEqual(mv.x, m.xx * v.x + m.xy * v.y + m.xz * v.z)
        self.assertEqual(mv.y, m.yx * v.x + m.yy * v.y + m.yz * v.z)
        self.assertEqual(mv.z, m.zx * v.x + m.zy * v.y + m.zz * v.z)
        z, o = Expr.float(0), Expr.float(1)
        identity = Matrix(o, z, z, z, o, z, z, z, o)
        self.assertEqual(identity * v, v)
        exchange_xy = Matrix(z, o, z, o, z, z, z, z, o)
        self.assertEqual(exchange_xy * v, Vector(v.y, v.x, v.z))
        exchange_xz = Matrix(z, z, o, z, o, z, o, z, z)
        self.assertEqual(exchange_xz * v, Vector(v.z, v.y, v.x))
        exchange_yz = Matrix(o, z, z, z, z, o, z, o, z)
        self.assertEqual(exchange_yz * v, Vector(v.x, v.z, v.y))

    def test_matrix_matrix_subtract(self):
        from mccode_antlr.instr.orientation import Matrix, Vector
        from mccode_antlr.common import Expr
        from numpy import random
        m1 = Matrix(*[Expr.float(random.rand()) for _ in range(9)])
        m2 = Matrix(*[Expr.float(random.rand()) for _ in range(9)])
        m3 = m1 - m2
        for i in range(9):
            self.assertEqual(m1[i] - m2[i], m3[i])
        m4 = m3 - m3
        v = Vector(*[Expr.float(random.rand()) for _ in range(3)])
        self.assertTrue((m4 * v).is_null)

    def test_vector_length(self):
        from mccode_antlr.instr.orientation import Vector
        from mccode_antlr.common import Expr
        from numpy import random, sqrt
        v = Vector(Expr.float(0), Expr.float(0), Expr.float(0))
        self.assertEqual(v.length(), Expr.float(0))
        self.assertTrue(v.is_null())

        v = Vector(Expr.float(2), Expr.float(0), Expr.float(0))
        self.assertEqual(v.length(), Expr.float(2))
        v = Vector(Expr.float(0), Expr.float(-9), Expr.float(0))
        self.assertEqual(v.length(), Expr.float(9))
        raw_v = [random.rand() for _ in range(3)]
        v = Vector(*[Expr.float(x) for x in raw_v])
        self.assertAlmostEqual(v.length(), Expr.float(sqrt(sum([x * x for x in raw_v]))))

    def test_rotation_matrix(self):
        from mccode_antlr.instr.orientation import Rotation, Angles, _rotation_angles_to_axes_coordinates
        from mccode_antlr.common import Expr
        from numpy import random, pi
        a = Angles(*[Expr.float(random.rand() * 2 * pi - pi) for _ in range(3)])
        while a.is_null():
            a = Angles(*[Expr.float(random.rand() * 2 * pi - pi) for _ in range(3)])
        axes, coordinates = _rotation_angles_to_axes_coordinates(a, degrees=False)
        r = Rotation(*axes)
        inv_r = r.inverse()
        one = r * inv_r
        identity = Rotation()
        for i in range(9):
            self.assertAlmostEqual(one[i], identity[i])

        zero = r - r
        for i in range(9):
            self.assertEqual(zero[i], Expr.float(0))

    def test_rotation_mccode_style(self):
        from mccode_antlr.instr.orientation import Rotation, Angles, Matrix
        from mccode_antlr.common import Expr
        o, z, p, m = [Expr.float(x) for x in (1, 0, 90, -90)]
        angles = {'zpz': Rotation.from_angles(Angles(z, p, z)),
                  'pzp': Rotation.from_angles(Angles(p, z, p)),
                  'zzp': Rotation.from_angles(Angles(z, z, p)),
                  'ppz': Rotation.from_angles(Angles(p, p, z)),
                  'mzp': Rotation.from_angles(Angles(m, z, p)),
                  'pzm': Rotation.from_angles(Angles(p, z, m)),
                  'mzm': Rotation.from_angles(Angles(m, z, m)),
                  'zpp': Rotation.from_angles(Angles(z, p, p)),
                  'zpm': Rotation.from_angles(Angles(z, p, m)),
                  'zmp': Rotation.from_angles(Angles(z, m, p)),
                  'zmm': Rotation.from_angles(Angles(z, m, m)),
                  'ppp': Rotation.from_angles(Angles(p, p, p)),
                  'ppm': Rotation.from_angles(Angles(p, p, m)),
                  'pmp': Rotation.from_angles(Angles(p, m, p)),
                  'pmm': Rotation.from_angles(Angles(p, m, m)),
                  'mpp': Rotation.from_angles(Angles(m, p, p)),
                  'mpm': Rotation.from_angles(Angles(m, p, m)),
                  'mmp': Rotation.from_angles(Angles(m, m, p)),
                  'mmm': Rotation.from_angles(Angles(m, m, m))}

        # Verified from McStas compile instrument _*_var.rotation_absolute via --trace output:
        mc = {'zpz': Rotation(z, z, -o, z, o, z, o, z, z),
              'pzz': Rotation(o, z, z, z, z, o, z, -o, z),
              'zzp': Rotation(z, o, z, -o, z, z, z, z, o),
              'pzp': Rotation(z, z, o, -o, z, z, z, -o, z),
              'mzp': Rotation(z, z, -o, -o, z, z, z, o, z),
              'pzm': Rotation(z, z, -o, o, z, z, z, -o, z),
              'mzm': Rotation(z, z, o, o, z, z, z, o, z),
              'zpp': Rotation(z, o, z, z, z, o, o, z, z),
              'zmp': Rotation(z, o, z, z, z, -o, -o, z, z),
              'zmm': Rotation(z, -o, z, z, z, o, -o, z, z),
              'zpm': Rotation(z, -o, z, z, z, -o, o, z, z),
              'ppz': Rotation(z, o, z, z, z, o, o, z, z),
              'ppp': Rotation(z, z, o, z, -o, z, o, z, z),
              'mpp': Rotation(z, z, -o, z, o, z, o, z, z),
              'pmp': Rotation(z, z, o, z, o, z, -o, z, z),
              'ppm': Rotation(z, z, -o, z, o, z, o, z, z),
              'mmp': Rotation(z, z, -o, z, -o, z, -o, z, z),
              'mpm': Rotation(z, z, o, z, -o, z, o, z, z),
              'pmm': Rotation(z, z, -o, z, -o, z, -o, z, z),
              'mmm': Rotation(z, z, o, z, o, z, -o, z, z)}

        # constructing the Rotation matrix may have small numerical errors, so we round to 14 decimal places
        # and compare to the zero Matrix
        for key in angles:
            self.assertEqual(round(abs(angles[key] - mc[key]), 14), Matrix())

    def test_seitz_multiply_identity(self):
        from numpy import random
        from mccode_antlr.common import Expr
        from mccode_antlr.instr.orientation import Seitz
        o, z = Expr.float(1), Expr.float(0)
        a = Seitz(o, z, z, z, z, o, z, z, z, z, o, z)
        b = a * a
        self.assertEqual(a, b)

        c = Seitz(*[Expr.float(random.rand()) for _ in range(12)])
        b = a * c
        self.assertEqual(b, c)

    def test_seitz_inverse(self):
        from numpy import random, pi
        from mccode_antlr.common import Expr
        from mccode_antlr.instr.orientation import Seitz, cos_value, sin_value
        o, z = Expr.float(1), Expr.float(0)
        a = Seitz(o, z, z, z, z, o, z, z, z, z, o, z)
        # Construct a matrix that *has* an inverse (use rotations to make life easier)
        ts = ([Expr.float(30/180*pi) for _ in range(3)], _random_angles_radian(),)
        for tx, ty, tz in ts:
            rx, ry, rz = _make_seitz_list(tx, ty, tz)
            b = rz * (ry * rx)
            c = b.inverse()
            d = c * b

            for i, j in zip(d, a):
                self.assertAlmostEqual(i, j)

            for i, j in zip(b, c.inverse()):
                self.assertAlmostEqual(i, j)

    def test_OrientationPart(self):
        from mccode_antlr.common import Expr
        from mccode_antlr.instr.orientation import Part, Vector
        op = Part()
        self.assertFalse(op.is_translation)
        self.assertFalse(op.is_rotation)
        self.assertTrue(op.is_identity)
        self.assertTrue(op.all_values)
        self.assertTrue(op.is_constant)

        axis, angle, unit = op.rotation_axis_angle
        self.assertEqual(unit, 'degrees')
        self.assertEqual(angle, Expr.float(0))
        # Since the angle is zero, the axis is degenerate and the returned value depends on details of the eigenvalue
        # solver, which we are not testing here.

        tx, ty, tz = _random_angles_degrees()
        while tx.is_zero or ty.is_zero or tz.is_zero:
            print(f'{tx=} {ty=} {tz=} must all be finite')
            tx, ty, tz = _random_angles_degrees()
        rx, ry, rz = _make_seitz_list(tx, ty, tz, degrees=True)

        # for individual rotations, it's easy to prove that axes and coordinates rotate opposite each other:
        opx = Part(rx)
        opy = Part(ry)
        opz = Part(rz)

        for o in (opx, opy, opz):
            if not o.is_rotation:
                print(f'This test fails despite finite {tx}, {ty} and {tz}? Why is {o} not a rotation?')
            self.assertTrue(o.is_rotation)
            self.assertFalse(o.is_translation)
            self.assertFalse(o.is_identity)
            self.assertTrue(o.all_values)
            self.assertTrue(o.is_constant)
            if not (o * o.inverse()).is_identity:
                # This less-robust identity test is needed because of possible numerical errors
                self.assertAlmostEqual((o * o.inverse()).axes.trace(), Expr.float(3),
                                       msg=f'{o} * {o.inverse()} = {o * o.inverse()} is not identity {tx=} {ty=} {tz=}')
            if not (o.inverse() * o).is_identity:
                self.assertAlmostEqual((o.inverse() * o).axes.trace(), Expr.float(3),
                                       msg=f'{o.inverse()} * {o} = {o.inverse() * o} is not identity {tx=} {ty=} {tz=}')

        vx = Vector(Expr.float(1), Expr.float(0), Expr.float(0))
        vy = Vector(Expr.float(0), Expr.float(1), Expr.float(0))
        vz = Vector(Expr.float(0), Expr.float(0), Expr.float(1))

        for o, t, v in [(opx, tx, vx), (opy, ty, vy), (opz, tz, vz)]:
            axis, angle, unit = o.rotation_axis_angle
            if (angle * t) < Expr.float(0):
                # this shouldn't happen since we've selected all vectors to be positive
                angle *= Expr.float(-1)
                axis = Vector(-axis.x, -axis.y, -axis.z)
            self.assertEqual(unit, 'degrees')
            self.assertAlmostEqual(angle, t)
            self.assertEqual(axis, v)

    def test_RotationParts(self):
        from mccode_antlr.common import Expr
        from mccode_antlr.instr.orientation import RotationPart, RotationX, RotationY, RotationZ, Angles, Vector

        rp = RotationPart()
        self.assertTrue(rp.is_constant)
        self.assertFalse(rp.is_translation)
        self.assertFalse(rp.is_rotation)
        self.assertTrue(rp.is_identity)
        self.assertEqual(rp.rotation_axis, Vector(Expr.float(0), Expr.float(0), Expr.float(0)))
        self.assertEqual(rp.position(), Vector(Expr.float(0), Expr.float(0), Expr.float(0)))

        # Since we use random angles, sample more of angle space to increase the chance of finding problems
        for _ in range(500):
            tx, ty, tz = _random_angles_degrees()
            # for individual rotations, it's easy to prove that axes and coordinates rotate opposite each other:
            rpx = RotationX(v=tx, degrees=True)
            self.assertTrue(rpx.is_constant)
            self.assertFalse(rpx.is_translation)
            self.assertTrue(rpx.is_rotation)
            self.assertFalse(rpx.is_identity)
            self.assertEqual(rpx.rotation_axis, Vector(Expr.float(1), Expr.float(0), Expr.float(0)))
            self.assertEqual(rpx.angles(), Angles(tx, Expr.float(0), Expr.float(0)))
            ex = rpx * rpx.inverse()
            if not ex.is_identity:
                # This less-robust identity test is needed because of possible numerical errors
                self.assertAlmostEqual(ex.axes.trace(), Expr.float(3),
                                       msg=f'{rpx} * {rpx.inverse()} = {ex} is not identity for {tx=}')

            rpy = RotationY(v=ty, degrees=True)
            self.assertTrue(rpy.is_constant)
            self.assertFalse(rpy.is_translation)
            self.assertTrue(rpy.is_rotation)
            self.assertFalse(rpy.is_identity)
            self.assertEqual(rpy.rotation_axis, Vector(Expr.float(0), Expr.float(1), Expr.float(0)))
            self.assertEqual(rpy.angles(), Angles(Expr.float(0), ty, Expr.float(0)))
            ey = rpy * rpy.inverse()
            if not ey.is_identity:
                self.assertAlmostEqual(ey.axes.trace(), Expr.float(3),
                                       msg=f'{rpy} * {rpy.inverse()} = {ey} is not identity for {ty=}')

            rpz = RotationZ(v=tz, degrees=True)
            self.assertTrue(rpz.is_constant)
            self.assertFalse(rpz.is_translation)
            self.assertTrue(rpz.is_rotation)
            self.assertFalse(rpz.is_identity)
            self.assertEqual(rpz.rotation_axis, Vector(Expr.float(0), Expr.float(0), Expr.float(1)))
            self.assertEqual(rpz.angles(), Angles(Expr.float(0), Expr.float(0), tz))
            ez = rpz * rpz.inverse()
            if not ez.is_identity:
                self.assertAlmostEqual(ez.axes.trace(), Expr.float(3),
                                       msg=f'{rpz} * {rpz.inverse()} = {ez} is not identity for {tz=}')

    def test_RotationPart_subclass_post_init_is_called(self):
        from mccode_antlr.instr.orientation import Rotation, RotationX, RotationY, RotationZ
        tx, ty, tz = _random_angles_degrees()
        while tx.is_zero or ty.is_zero or tz.is_zero:
            print(f'{tx=} {ty=} {tz=} must all be finite')
            tx, ty, tz = _random_angles_degrees()

        rpx = RotationX(v=tx, degrees=True)
        self.assertFalse(rpx.is_identity)
        self.assertNotEqual(rpx.rotation(), Rotation())
        rpy = RotationY(v=ty, degrees=True)
        self.assertFalse(rpy.is_identity)
        self.assertNotEqual(rpy.rotation(), Rotation())
        rpz = RotationZ(v=tz, degrees=True)
        self.assertFalse(rpz.is_identity)
        self.assertNotEqual(rpz.rotation(), Rotation())

    def test_Vector(self):
        from mccode_antlr.instr.orientation import Vector
        from mccode_antlr.common import Expr
        tx = Vector(Expr.float(1), Expr.float(0), Expr.float(0))
        ty = Vector(Expr.float(0), Expr.float(1), Expr.float(0))
        tz = Vector(Expr.float(0), Expr.float(0), Expr.float(1))

        cxy = tx * ty.cross_matrix()
        self.assertEqual(cxy, tz)
        czx = tz * tx.cross_matrix()
        self.assertEqual(czx, ty)
        cyz = ty * tz.cross_matrix()
        self.assertEqual(cyz, tx)

    def test_TranslationPart(self):
        from mccode_antlr.instr.orientation import TranslationPart, Vector
        t = _random_vector(0.1, 1.0)
        tp = TranslationPart(v=t)
        self.assertTrue(tp.is_translation)
        self.assertFalse(tp.is_rotation)
        self.assertFalse(tp.is_identity)
        self.assertEqual(tp.position(), t)
        self.assertTrue((tp * tp.inverse()).is_identity)

    def test_OrientationParts(self):
        from mccode_antlr.common import Expr
        from mccode_antlr.instr.orientation import TranslationPart, Vector
        from mccode_antlr.instr.orientation import RotationX, RotationY, RotationZ
        from mccode_antlr.instr.orientation import Parts

        tx = TranslationPart(v=Vector(Expr.float(0.1), Expr.float(0), Expr.float(0)))
        ty = TranslationPart(v=Vector(Expr.float(0.), Expr.float(0.2), Expr.float(0)))
        tz = TranslationPart(v=Vector(Expr.float(0.), Expr.float(0), Expr.float(0.3)))

        op = Parts((tx, ty, tz)).reduce()
        self.assertEqual(len(op.stack()), 1)
        t = op.stack()[0]
        self.assertEqual(t, TranslationPart(v=Vector(Expr.float(0.1), Expr.float(0.2), Expr.float(0.3))))

        rx = RotationX(v=Expr.float(30), degrees=True)
        ry = RotationY(v=Expr.float(45), degrees=True)
        rz = RotationZ(v=Expr.float(60), degrees=True)

        rx2 = Parts((rx, rx)).reduce()
        self.assertEqual(len(rx2.stack()), 1)
        rx2 = rx2.stack()[0]
        rx60 = RotationX(v=Expr.float(60), degrees=True)
        for a, b in zip(rx2.seitz(), rx60.seitz()):
            self.assertAlmostEqual(a, b)

        for a, b in zip((Parts((rx,)) + Parts((rx,))).reduce().stack()[0].seitz(), rx2.seitz()):
            self.assertEqual(a, b)

        ry2 = Parts((ry, ry)).reduce()
        self.assertEqual(len(ry2.stack()), 1)
        ry2 = ry2.stack()[0]
        ry90 = RotationY(v=Expr.float(90), degrees=True)
        for a, b in zip(ry2.seitz(), ry90.seitz()):
            self.assertAlmostEqual(a, b)

        rz2 = Parts((rz, rz)).reduce()
        self.assertEqual(len(rz2.stack()), 1)
        rz2 = rz2.stack()[0]
        rz120 = RotationZ(v=Expr.float(120), degrees=True)
        for a, b in zip(rz2.seitz(), rz120.seitz()):
            self.assertAlmostEqual(a, b)

        rxy = Parts((rx, ry)).reduce().stack()[0]
        ryx = Parts((ry, rx)).reduce().stack()[0]
        for j in ('xx', 'yy', 'zz'):
            self.assertAlmostEqual(getattr(rxy.seitz(), j), getattr(ryx.seitz(), j))
        for i, j in [('xy', 'yx')]:
            self.assertAlmostEqual(getattr(rxy.seitz(), i), getattr(ryx.seitz(), j))
        for i, j in [('xz', 'zx'), ('yz', 'zy')]:
            self.assertAlmostEqual(getattr(rxy.seitz(), i), -getattr(ryx.seitz(), j))

        trx = Parts((t, rx)).reduce().stack()[0]
        rxt = Parts((rx, t)).reduce().stack()[0]
        for j in ('xx', 'yy', 'zz', 'xy', 'yx', 'xz', 'zx', 'yz', 'zy', 'xt'):
            self.assertAlmostEqual(getattr(trx.seitz(), j), getattr(rxt.seitz(), j))
        for j in ('yt', 'zt'):
            self.assertNotAlmostEqual(getattr(trx.seitz(), j), getattr(rxt.seitz(), j))

        trz = Parts((t, rz)).reduce().stack()[0]
        rzt = Parts((rz, t)).reduce().stack()[0]
        for j in ('xx', 'yy', 'zz', 'xy', 'yx', 'xz', 'zx', 'yz', 'zy', 'zt'):
            self.assertAlmostEqual(getattr(trz.seitz(), j), getattr(rzt.seitz(), j))
        for j in ('xt', 'yt'):
            self.assertNotAlmostEqual(getattr(trz.seitz(), j), getattr(rzt.seitz(), j))

        trz2 = Parts() + t + rz
        self.assertTrue(isinstance(trz2, Parts))
        self.assertEqual(len(trz2.stack()), 2)
        trz2 = trz2.reduce().stack()[0]
        self.assertEqual(trz2, trz)

    def test_OrientationParts_from_at_rotated(self):
        from mccode_antlr.instr.orientation import Parts, Seitz, RotationX, RotationY, RotationZ, TranslationPart, Angles
        from mccode_antlr.common import Expr
        tx, ty, tz = _random_angles_degrees()
        rx, ry, rz = _make_seitz_list(tx, ty, tz, degrees=True)
        t = _random_vector(0.1, 1.0)
        op = Parts.from_at_rotated(t, Angles(tx, ty, tz), degrees=True)
        self.assertEqual(len(op.stack()), 4)
        self.assertEqual(op.stack()[0], RotationX(v=Expr.float(tx), degrees=True))
        self.assertEqual(op.stack()[1], RotationY(v=Expr.float(ty), degrees=True))
        self.assertEqual(op.stack()[2], RotationZ(v=Expr.float(tz), degrees=True))
        self.assertEqual(op.stack()[3], TranslationPart(v=t))

        inv_op = op.inverse()
        self.assertEqual(len(inv_op.stack()), 4)
        self.assertEqual(inv_op.stack()[1], RotationZ(v=Expr.float(-tz), degrees=True))
        self.assertEqual(inv_op.stack()[2], RotationY(v=Expr.float(-ty), degrees=True))
        self.assertEqual(inv_op.stack()[3], RotationX(v=Expr.float(-tx), degrees=True))

        full = (op + inv_op).reduce()
        resolved = (op + inv_op).resolve()
        self.assertEqual(len(full.stack()), 1)
        full = full.stack()[0]
        self.assertEqual(full, resolved)
        full = full.seitz()
        identity = Seitz()
        for i in range(9):
            self.assertAlmostEqual(full[i], identity[i])

    def test_DependentOrientation_no_rotations(self):
        from mccode_antlr.common import Expr
        from mccode_antlr.instr.orientation import TranslationPart, Vector, Angles
        from mccode_antlr.instr.orientation import Parts, Orient

        v123 = Vector(Expr.float(0.1), Expr.float(0.2), Expr.float(0.3))
        v321 = Vector(Expr.float(0.3), Expr.float(0.2), Expr.float(0.1))
        t123 = TranslationPart(v=v123)
        t321 = TranslationPart(v=v321)

        a000 = Angles(Expr.float(0), Expr.float(0), Expr.float(0))

        dop = Orient.from_dependent_orientation(None, v123, a000)
        self.assertEqual(dop.position(), v123)

        d2 = Orient.from_dependent_orientation(dop, v321, a000)
        t444 = Parts((t123, t321)).reduce().stack()[0]
        self.assertEqual(d2.position(), t444.position())
        self.assertEqual(d2.angles(), a000)

    # def test_DependentOrientation_with_rotations(self):
    #     from mccode_antlr.common import Expr
    #     from mccode_antlr.instr.orientation import TranslationPart, Vector, Angles
    #     from mccode_antlr.instr.orientation import RotationX, RotationY, RotationZ
    #     from mccode_antlr.instr.orientation import OrientParts, Orient
    #
    #     tx, ty, tz = _random_angles_degrees()
    #     f = Expr.float
    #
    #     for a in (Angles(tx, f(0), f(0)), Angles(f(0), ty, f(0)), Angles(f(0), f(0), tz)):
    #         v = _random_vector(0.1, 1.0)
    #         dop = Orient.from_dependent_orientation(None, v, a)
    #         self.assertEqual(dop.position(), v)
    #         self.assertEqual(dop.angles(), a)
    #         print(dop)
    #
    #     # Rotate 90 degrees around z axis
    #     z90 = Angles(f(0), f(0), f(90))
    #     # Then rotate 90 degrees around x axis
    #     x90 = Angles(f(90), f(0), f(0))
    #     # This permutes the axes x'=y, y'=z, z'=x
    #     xyz = Orient()
    #     inter = Orient.from_dependent_orientation(xyz, Vector(f(0), f(0), f(0)), z90)
    #     yzx = Orient.from_dependent_orientation(inter, Vector(f(0), f(0), f(0)), x90)
    #
    #     self.assertEqual(yzx._position.resolve().seitz(), yzx._rotation.resolve().seitz())
    #
    #     self.assertNotEqual(yzx.angles(), Angles(f(90), f(0), f(90)))
    #     self.assertEqual(yzx.angles(), Angles(f(180), f(90), f(-90)))
    #
    #
    #     #
    #     # d2 = DependentOrientation.from_dependent_orientation(dop, v321, a000)
    #     # t444 = OrientationParts((t123, t321)).reduce().stack()[0]
    #     # self.assertEqual(d2.position(), t444.position())
    #     # self.assertEqual(d2.angles(), a000)

    def test_DependentOrientation_simple(self):
        from mccode_antlr.common import Expr
        from mccode_antlr.instr.orientation import Vector, Angles, Orient

        v_z1 = Vector(Expr.float(0), Expr.float(0), Expr.float(1))
        v_x1 = Vector(Expr.float(1), Expr.float(0), Expr.float(0))
        v_y1 = Vector(Expr.float(0), Expr.float(1), Expr.float(0))
        v_0 = Vector(Expr.float(0), Expr.float(0), Expr.float(0))
        a_z90 = Angles(Expr.float(0), Expr.float(0), Expr.float(90))
        a_0 = Angles(Expr.float(0), Expr.float(0), Expr.float(0))

        origin = Orient()
        one = Orient.from_dependent_orientation(origin, v_0, a_z90)
        two = Orient.from_dependent_orientation(one, v_z1, a_z90)
        three = Orient.from_dependent_orientation(two, v_z1, a_0)
        four = Orient.from_dependent_orientation(three, v_x1, a_0)

        self.assertEqual(origin.position(), v_0)
        self.assertEqual(origin.angles(), a_0)
        self.assertEqual(one.position(), v_0)
        self.assertEqual(one.angles(), a_z90)
        self.assertEqual(two.position(), v_z1)
        self.assertEqual(two.angles(), a_z90 * Expr.float(2))
        self.assertEqual(three.position(), v_z1 + v_z1)
        self.assertEqual(four.position(), three.position() - v_x1)

    def test_DepedentOrientation_triple_axis(self):
        from math import sin, cos
        from mccode_antlr.common import Expr, unary_expr
        from mccode_antlr.instr.orientation import Vector, Angles, Orient, Rotation, Parts, RotationY

        def ri(i):
            return Angles(Expr.float(0), Expr.id(f'a{i}'), Expr.float(0))

        r1, r2, r3, r4, r5, r6 = [ri(i) for i in range(1, 7)]
        v1 = Vector(Expr.float(0), Expr.float(0), Expr.float(1))
        a0 = Angles(Expr.float(0), Expr.float(0), Expr.float(0))
        v0 = Vector(Expr.float(0), Expr.float(0), Expr.float(0))

        origin = Orient()
        monochromator = Orient.from_dependent_orientation(origin, v1, r1)
        mono_arm = Orient.from_dependent_orientations(monochromator, v0, origin, r2)
        sample = Orient.from_dependent_orientation(mono_arm, v1, r3)
        sample_arm = Orient.from_dependent_orientations(sample, v0, mono_arm, r4)
        analyzer = Orient.from_dependent_orientation(sample_arm, v1, r5)
        ana_arm = Orient.from_dependent_orientations(analyzer, v0, sample_arm, r6)
        detector = Orient.from_dependent_orientation(ana_arm, v1, a0)

        self.assertEqual(origin.position(), v0)
        self.assertEqual(origin.angles(), a0)
        self.assertEqual(monochromator.position(), v1)
        self.assertEqual(monochromator.angles(), r1)
        self.assertEqual(mono_arm.position(), v1)
        self.assertEqual(mono_arm.angles(), r2)
        #self.assertEqual(sample.position(), v1 +v1)

        ai = [Expr.id(f'a{i}') for i in range(7)]
        sa = [unary_expr(sin, 'sin', a) for a in ai]
        ca = [unary_expr(cos, 'cos', a) for a in ai]
        sample_position = Vector(sa[2], Expr.float(0.0), Expr.float(1) + ca[2])
        sample_analyzer_vector = Vector(ca[2] * sa[4] + sa[2] * ca[4], Expr.float(0), - sa[2] * sa[4] + ca[4] * ca[2])

        self.assertEqual(sample.position(), sample_position)
        # Matching up expression trees is hard -- so use this special hand-checked equivalent matrix
        er = Rotation(ca[3] * ca[2] + -sa[3] * sa[2], Expr.float(0), ca[3] * -sa[2] + -sa[3] * ca[2],
                      Expr.float(0), Expr.float(1), Expr.float(0),
                      sa[3] * ca[2] + ca[3] * sa[2], Expr.float(0), sa[3] * -sa[2] + ca[3] * ca[2])
        self.assertEqual(sample.rotation(), er)

        self.assertEqual(sample_arm.position(), sample_position)
        er = Rotation(ca[4] * ca[2] + -sa[4] * sa[2], Expr.float(0), ca[4] * -sa[2] + -sa[4] * ca[2],
                      Expr.float(0), Expr.float(1), Expr.float(0),
                      sa[4] * ca[2] + ca[4] * sa[2], Expr.float(0), sa[4] * -sa[2] + ca[4] * ca[2])
        self.assertEqual(sample_arm.rotation(), er)

        self.assertEqual(analyzer.position(), sample_position + sample_analyzer_vector)

        self.assertEqual(analyzer.rotation_parts(),
                         Parts((RotationY(v=ai[2]), RotationY(v=ai[4]), RotationY(v=ai[5]))))

        self.assertEqual(ana_arm.rotation_parts(),
                         Parts((RotationY(v=ai[2]), RotationY(v=ai[4]), RotationY(v=ai[6]))))

        # testing the detector position is more of the same, but more complicated, than the analyzer position
        # Since it doesn't test much, skip it for the moment
        self.assertEqual(ana_arm.rotation(), detector.rotation())
